# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.0.0 / 2024.12.15
# Sphinx OK.

"""Image processing helpers."""

import numpy as np

from . import params

class Container:
  """An empty container class."""
  pass # An empty container class.

def failsafe_divide(A, B):
  """Return A/B, ignoring errors (division by zero, ...).

  Args:
    A (numpy.ndarray): The numerator array.
    B (numpy.ndarray): The denominator array.

  Returns:
    numpy.ndarray: The element-wise division A/B.
  """
  status = np.seterr(all = "ignore")
  C = np.divide(A, B)
  np.seterr(**status)
  return C

def scale_pixels(image, source, target, cutoff = params.IMGTOL):
  """Scale all pixels of the image by the ratio target/source. Wherever abs(source) < cutoff, set all channels to target.

  Args:
    image (numpy.ndarray): The input image.
    source (numpy.ndarray): The source values for scaling (must be the same size as the input image).
    target (numpy.ndarray): The target values for scaling (must be the same size as the input image).
    cutoff (float, optional): Threshold for scaling. Defaults to `equimage.params.IMGTOL`.

  Returns:
    numpy.ndarray: The scaled image.
  """
  return np.where(abs(source) > cutoff, failsafe_divide(image*target, source), target)

def lookup(x, xlut, ylut, slut, nlut):
  """Linearly interpolate y = f(x) between the values ylut = f(xlut) of an evenly spaced look-up table.

  Args:
    x (float): The input value for interpolation.
    xlut (numpy.ndarray): The x values of the look-up table (must be evenly spaced).
    ylut (numpy.ndarray): The y values of the look-up table ylut = f(xlut).
    slut (numpy.ndarray): The slopes (ylut[1:]-ylut[:-1])/(xlut[1:]-xlut[:-1]) used for linear interpolation between the elements of ylut.
    nlut (int): The number of elements in the look-up table.

  Returns:
    float: The interpolated value y = f(x).
  """
  i = np.clip(np.int32(np.floor((nlut-1)*(x-xlut[0])/(xlut[-1]-xlut[0]))), 0, nlut-2)
  return slut[i]*(x-xlut[i])+ylut[i]



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.0.0 / 2024.12.15
# Sphinx OK.

"""Image processing tools."""

# Import everything.

from .image import Image
from .params import get_RGB_luma, set_RGB_luma
from .image_utils import is_valid_image, clip, blend
from .image_colorspaces import value, saturation, luma
from .image_stretch import mts, ghs
from .image_masks import threshold_mask
from .image_io import load_image, save_image



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.0.0 / 2024.12.15
# Sphinx OK.

"""Image utils."""

import numpy as np

from . import params
from . import helpers

#####################
# Image validation. #
#####################

def is_valid_image(image):
  """Return True if the input is a valid image candidate, False otherwise.

  Args:
    image (numpy.ndarray): The image candidate.

  Returns:
    bool: True if the input is a valid image candidate, False otherwise.
  """
  if not issubclass(type(image), np.ndarray): return False
  if image.ndim == 3:
    if image.shape[0] not in [1, 3]: return False
  elif image.ndim != 2:
    return False
  if image.dtype not in [np.float32, np.float64]: return False

##########################
# Image transformations. #
##########################

def clip(image, vmin = 0., vmax = 1.):
  """Clip the input image in the range [vmin, vmax].

  Args:
    image (numpy.ndarray): The input image.
    vmin (float, optional): The lower clip bound (default 0).
    vmax (float, optional): The upper clip bound (default 1).

  Returns:
    numpy.ndarray: The clipped image.
  """
  return np.clip(image, vmin, vmax)

def blend(image1, image2, mixing):
  """Blend two images.

  Returns image1*(1-mixing)+image2*mixing.

  Args:
    image1 (numpy.ndarray): The first image.
    image2 (numpy.ndarray): The second image.
    mixing (float or numpy.ndarray for pixel-dependent mixing): The mixing coefficient(s).

  Returns:
    numpy.ndarray: The blended image image1*(1-mixing)+image2*mixing.
  """
  return image1*(1.-mixing)+image2*mixing

#####################################
# For inclusion in the Image class. #
#####################################

class Mixin:
  """To be included in the Image class."""

  ##################
  # Image queries. #
  ##################

  def is_out_of_range(self):
    """Return True if the image is out-of-range (data < 0 or > 1 in any channel), False otherwise.

    Returns:
      bool: True if the image is out-of-range, False otherwise.
    """
    return np.any(self.image < -params.IMGTOL) or np.any(self.image > 1.+params.IMGTOL)

  ##############
  # Templates. #
  ##############

  def empty(self):
    """Return an empty image with same size as the object.

    Returns:
      Image: An empty image with the same size as self.
    """
    return np.empty_like(self)

  def black(self):
    """Return a black image with same size as the object.

    Returns:
      Image: An black image with the same size as self.
    """
    return np.zeros_like(self)

  ##############################
  # Clipping & scaling pixels. #
  ##############################

  def clip(self, vmin = 0., vmax = 1.):
    """Clip the image in the range [vmin, vmax].

    Args:
      vmin (float, optional): The lower clip bound (default 0).
      vmax (float, optional): The upper clip bound (default 1).

    Returns:
      Image: The clipped image.
    """
    return np.clip(self, vmin, vmax)

  def scale_pixels(self, source, target, cutoff = params.IMGTOL):
    """Scale all pixels of the image by the ratio target/source. Wherever abs(source) < cutoff, set all channels to target.

    Args:
      source (np.arrray): The source values for scaling (must be the same size as the image).
      target (np.arrray): The target values for scaling (must be the same size as the image).
      cutoff (float, optional): Threshold for scaling. Defaults to `equimage.params.IMGTOL`.

    Returns:
      Image: The scaled image.
    """
    return self.newImage(helpers.scale_pixels(self.image, np.asarray(source), np.asarray(target), cutoff))

  #############
  # Blending. #
  #############

  def blend(self, image, mixing):
    """Blend with the input image.

    Returns self*(1-mixing)+image*mixing.
    The images must share the same shape, color space and color model.

    Args:
      image (Image): The image to blend with.
      mixing (float or numpy.ndarray for pixel-dependent mixing): The mixing coefficient(s).

    Returns:
      Image: The blended image self*(1-mixing)+image*mixing.
    """
    if self.get_shape() != image.get_shape():
      raise ValueError("Error, the images must share the same size & number of channels.")
    if self.colorspace != image.colorspace:
      raise ValueError("Error, the images must share the same color space !")
    if self.colormodel != image.colormodel:
      raise ValueError("Error, the images must share the same color model !")
    return self.newImage(self.image*(1.-mixing)+image.image*mixing)



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.0.0 / 2024.12.15
# Sphinx OK.

"""eQuimage lab."""

import os
os.environ["LANGUAGE"] = "en"
import inspect

__version__ = "1.0.0"
__packagepath__ = os.path.dirname(inspect.getabsfile(inspect.currentframe()))

# Import everything.

from .equimage import *
from .jupyter import *

ruler = "######################################"+len(__version__)*"#"
print(ruler)
print(f"# Welcome to eQuimageLab version {__version__}... #")
print(ruler)
# print(f"Documentation available at: file://{__packagepath__}/docs/build/html/")
print("Documentation available at: https://astro.ymniquet.fr/codes/equimagelab/docs/")
print("The module numpy is loaded as np.")

set_RGB_luma("human")



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.0.0 / 2024.12.15
# Sphinx OK.

"""Histogram stretch."""

import numpy as np

from . import stretchfunctions as stf

######################
# Stretch functions. #
######################

def ghs(image, lnD1, b, SYP, SPP = 0., HPP = 1.):
  """Apply a generalized hyperbolic stretch function to the input image.

  For details about generalized hyperbolic stretches, see: https://ghsastro.co.uk/.
  This function clips the input image in the [0, 1] range before stretching.

  Args:
    image (numpy.ndarray): The input image.
    logD1 (float): The global stretch factor ln(D+1) (must be >= 0).
    b (float): The local stretch factor.
    SYP (float): The symmetry point (expected in [0, 1]).
    SPP (float, optional): The shadow protection point (default 0; expected in [0, SYP]).
    HPP (float, optional): The highlight protection point (default 1; expected in [SYP, 1]).
    inverse (bool): Return the inverse stretch function if True.

  Returns:
    numpy.ndarray: The stretched image.
  """
  return stf.ghyperbolic_stretch_function(image, lnD1, b, SYP, SPP, HPP, False)

def mts(image, midtone):
  """Apply a midtone stretch function to the input image.

  The midtone stretch function is defined as:

    f(x) = (midtone-1)*x/((2*midtone-1)*x-midtone)

  In particular, f(0) = 0, f(midtone) = 0.5 and f(1) = 1.

  Args:
    image (numpy.ndarray): The input image.
    midtone (float): The midtone level (expected in ]0, 1[).

  Returns:
    numpy.ndarray: The stretched image.
  """
  return stf.midtone_stretch_function(image, midtone, False)

#####################################
# For inclusion in the Image class. #
#####################################

class Mixin:
  """To be included in the Image class."""

  def set_black_point(self, shadow, channels = "", trans = True):
    """Set the black (shadow) level in selected channels of the image.

    The selected channels are clipped below shadow and linearly stretched to map [shadow, 1] onto [0, 1].
    The output, stretched image channels therefore fit in the [0, infty[ range.

    Args:
      shadow (float): The black (shadow) level (expected < 1).
      channels (str, optional): The selected channels:

        - An empty string (default): Apply the operation to all channels (RGB, HSV and grayscale images).
        - A combination of "1", "2", "3" (or equivalently "R", "G", "B" for RGB images): Apply the
          operation to the first/second/third channel (RGB, HSV and grayscale images).
        - "V": Apply the operation to the HSV value (RGB, HSV and and grayscale images).
        - "S": Apply the operation to the HSV saturation (RGB and HSV images).
        - "L": Apply the operation to the luma (RGB and grayscale images).
        - "Ls": Apply the operation to the luma, with highlights protection by desaturation
          (after the operation, the out-of-range pixels are desaturated at constant luma).
        - "Lb": Apply the operation to the luma, with highlights protection by blending
          (after the operation, the out-of-range pixels are blended with channels = "RGB").

      trans(bool, optional): If True (default), embed the transormation in the output image as
        output.trans (see Image.apply_channels).

    Returns:
      Image: The processed image.
    """
    if shadow > .9999: raise ValuerError("Error, shadow must be <= 0.9999.")
    output = self.apply_channels(lambda channel: stf.shadow_stretch_function(channel, shadow), channels, trans = trans)
    if trans: output.trans.xticks = [shadow]
    return output

  def clip_shadow_highlight(self, shadow, highlight, channels = "", trans = True):
    """Clip shadows and highlights in selected channels of the image.

    The selected channels are clipped below shadow and above highlight and linearly stretched
    to map [shadow, highlight] onto [0, 1].
    The output, stretched channels levels therefore fit in the [0, 1] range.

    Args:
      shadow (float): The shadow level (expected < 1).
      highlight (float): The highlight level (expected > shadow).
      channels (str, optional): The selected channels:

        - An empty string (default): Apply the operation to all channels (RGB, HSV and grayscale images).
        - A combination of "1", "2", "3" (or equivalently "R", "G", "B" for RGB images): Apply the
          operation to the first/second/third channel (RGB, HSV and grayscale images).
        - "V": Apply the operation to the HSV value (RGB, HSV and and grayscale images).
        - "S": Apply the operation to the HSV saturation (RGB and HSV images).
        - "L": Apply the operation to the luma (RGB and grayscale images).
        - "Ls": Apply the operation to the luma, with highlights protection by desaturation
          (after the operation, the out-of-range pixels are desaturated at constant luma).
        - "Lb": Apply the operation to the luma, with highlights protection by blending
          (after the operation, the out-of-range pixels are blended with channels = "RGB").

      trans(bool, optional): If True (default), embed the transormation in the output image as
        output.trans (see Image.apply_channels).

    Returns:
      Image: The processed image.
    """
    if shadow > .9999: raise ValuerError("Error, shadow must be <= 0.9999.")
    if highlight-shadow < .0001: raise ValuerError("Error, highlight-shadow must be >= 0.0001.")
    output = self.apply_channels(lambda channel: stf.shadow_highlight_stretch_function(channel, shadow, highlight), channels, trans = trans)
    if trans: output.trans.xticks = [shadow, highlight]
    return output

  def set_dynamic_range(self, fr, to, channels = "", trans = True):
    """Set the dynamic range of selected channels of the image.

    The selected channels are linearly stretched to map [fr[0], fr[1]] onto [to[0], to[1]],
    and clipped in the [to[0], to[1]] range.

    Args:
      fr (a tuple or list of two floats such that fr[1] > fr[0]): The input range.
      to (a tuple or list of two floats such that to[1] > to[0]): The output range.
      channels (str, optional): The selected channels:

        - An empty string (default): Apply the operation to all channels (RGB, HSV and grayscale images).
        - A combination of "1", "2", "3" (or equivalently "R", "G", "B" for RGB images): Apply the
          operation to the first/second/third channel (RGB, HSV and grayscale images).
        - "V": Apply the operation to the HSV value (RGB, HSV and and grayscale images).
        - "S": Apply the operation to the HSV saturation (RGB and HSV images).
        - "L": Apply the operation to the luma (RGB and grayscale images).
        - "Ls": Apply the operation to the luma, with highlights protection by desaturation
          (after the operation, the out-of-range pixels are desaturated at constant luma).
        - "Lb": Apply the operation to the luma, with highlights protection by blending
          (after the operation, the out-of-range pixels are blended with channels = "RGB").

      trans(bool, optional): If True (default), embed the transormation in the output image as
        output.trans (see Image.apply_channels).

    Returns:
      Image: The processed image.
    """
    if fr[1]-fr[0] < 0.0001: raise ValueError("Error, fr[1]-fr[0] must be >= 0.0001 !")
    if to[1]-to[0] < 0.0001: raise ValueError("Error, to[1]-to[0] must be >= 0.0001 !")
    return self.apply_channels(lambda channel: stf.dynamic_range_stretch_function(channel, fr, to), channels, trans = trans)

  def asinh_stretch(self, stretch, channels = "", trans = True):
    """Apply an arcsinh stretch to selected channels of the image.

    The arcsinh stretch function f is applied to the selected channels:

      f(x) = arcsinh(stretch*x)/arcsinh(stretch)

    Args:
      stretch (float): The stretch factor (must be >= 0).
      channels (str, optional): The selected channels:

        - An empty string (default): Apply the operation to all channels (RGB, HSV and grayscale images).
        - A combination of "1", "2", "3" (or equivalently "R", "G", "B" for RGB images): Apply the
          operation to the first/second/third channel (RGB, HSV and grayscale images).
        - "V": Apply the operation to the HSV value (RGB, HSV and and grayscale images).
        - "S": Apply the operation to the HSV saturation (RGB and HSV images).
        - "L": Apply the operation to the luma (RGB and grayscale images).
        - "Ls": Apply the operation to the luma, with highlights protection by desaturation
          (after the operation, the out-of-range pixels are desaturated at constant luma).
        - "Lb": Apply the operation to the luma, with highlights protection by blending
          (after the operation, the out-of-range pixels are blended with channels = "RGB").

      trans(bool, optional): If True (default), embed the transormation in the output image as
        output.trans (see Image.apply_channels).

    Returns:
      Image: The stretched image.
    """
    if stretch < 0.: raise ValueError("Error, stretch must be >= 0.")
    return self.apply_channels(lambda channel: stf.asinh_stretch_function(channel, stretch), channels, trans = trans)

  def ghyperbolic_stretch(self, lnD1, b, SYP, SPP = 0., HPP = 1., inverse = False, channels = "", trans = True):
    """Apply a generalized hyperbolic stretch to selected channels of the image.

    For details about generalized hyperbolic stretches, see: https://ghsastro.co.uk/.
    This function clips the selected channels in the [0, 1] range before stretching.

    Args:
      logD1 (float): The global stretch factor ln(D+1) (must be >= 0).
      b (float): The local stretch factor.
      SYP (float): The symmetry point (will be clipped in the [0, 1] range).
      SPP (float, optional): The shadow protection point (default 0, will be clipped in the [0, SYP] range).
      HPP (float, optional): The highlight protection point (default 1, will be clipped in the [SYP, 1] range).
      inverse (bool, optional): Return the inverse stretch if True (default False).
      channels (str, optional): The selected channels:

        - An empty string (default): Apply the operation to all channels (RGB, HSV and grayscale images).
        - A combination of "1", "2", "3" (or equivalently "R", "G", "B" for RGB images): Apply the
          operation to the first/second/third channel (RGB, HSV and grayscale images).
        - "V": Apply the operation to the HSV value (RGB, HSV and and grayscale images).
        - "S": Apply the operation to the HSV saturation (RGB and HSV images).
        - "L": Apply the operation to the luma (RGB and grayscale images).
        - "Ls": Apply the operation to the luma, with highlights protection by desaturation
          (after the operation, the out-of-range pixels are desaturated at constant luma).
        - "Lb": Apply the operation to the luma, with highlights protection by blending
          (after the operation, the out-of-range pixels are blended with channels = "RGB").

      trans(bool, optional): If True (default), embed the transormation in the output image as
        output.trans (see Image.apply_channels).

    Returns:
      numpy.ndarray: The stretched image.
    """
    if lnD1 < 0.: raise ValueError("Error, lnD1 must be >= 0.")
    if SYP < 0.:
      SYP = 0.
      print("Warning, changed SYP = 0 !")
    if SYP > 1.:
      SYP = 1.
      print("Warning, changed SYP = 1 !")
    if SPP < 0.:
      SPP = 0.
      print("Warning, changed SPP = 0 !")
    if SPP > SYP:
      SPP = SYP
      print("Warning, changed SPP = SYP !")
    if HPP < SYP:
      HPP = SYP
      print("Warning, changed HPP = SYP !")
    if HPP > 1.:
      HPP = 1.
      print("Warning, changed HPP = 1 !")
    output = self.apply_channels(lambda channel: stf.ghyperbolic_stretch_function(channel, lnD1, b, SYP, SPP, HPP, inverse), channels, trans = trans)
    if trans: output.trans.xticks = [SPP, SYP, HPP]
    return output

  def midtone_stretch(self, midtone, inverse = False, channels = "", trans = True):
    """Apply a midtone stretch to selected channels of the image.

    The midtone stretch function f is applied to the selected channels:

      f(x) = (midtone-1)*x/((2*midtone-1)*x-midtone)

    In particular, f(0) = 0, f(midtone) = 0.5 and f(1) = 1.

    Args:
      midtone (float): The midtone level (expected in ]0, 1[).
      inverse (bool, optional): Return the inverse stretch if True (default False).
      channels (str, optional): The selected channels:

        - An empty string (default): Apply the operation to all channels (RGB, HSV and grayscale images).
        - A combination of "1", "2", "3" (or equivalently "R", "G", "B" for RGB images): Apply the
          operation to the first/second/third channel (RGB, HSV and grayscale images).
        - "V": Apply the operation to the HSV value (RGB, HSV and and grayscale images).
        - "S": Apply the operation to the HSV saturation (RGB and HSV images).
        - "L": Apply the operation to the luma (RGB and grayscale images).
        - "Ls": Apply the operation to the luma, with highlights protection by desaturation
          (after the operation, the out-of-range pixels are desaturated at constant luma).
        - "Lb": Apply the operation to the luma, with highlights protection by blending
          (after the operation, the out-of-range pixels are blended with channels = "RGB").

      trans(bool, optional): If True (default), embed the transormation in the output image as
        output.trans (see Image.apply_channels).

    Returns:
      Image: The stretched image.
    """
    if midtone < .0001 or midtone >= .9999: raise ValueError("Error, midtone must be >= 0.0001 and <= 0.9999.")
    output = self.apply_channels(lambda channel: stf.midtone_stretch_function(channel, midtone, inverse), channels, trans = trans)
    if trans: output.trans.xticks = [midtone]
    return output

  def gamma_stretch(self, gamma, channels = "", trans = True):
    """Apply a power law stretch (gamma correction) to selected channels of the image.

    The gamma stretch function f is applied to the selected channels:

      f(x) = x**gamma

    This function clips the image below 0 before stretching.

    Args:
      gamma (float): The stretch exponent (must be > 0).
      channels (str, optional): The selected channels:

        - An empty string (default): Apply the operation to all channels (RGB, HSV and grayscale images).
        - A combination of "1", "2", "3" (or equivalently "R", "G", "B" for RGB images): Apply the
          operation to the first/second/third channel (RGB, HSV and grayscale images).
        - "V": Apply the operation to the HSV value (RGB, HSV and and grayscale images).
        - "S": Apply the operation to the HSV saturation (RGB and HSV images).
        - "L": Apply the operation to the luma (RGB and grayscale images).
        - "Ls": Apply the operation to the luma, with highlights protection by desaturation
          (after the operation, the out-of-range pixels are desaturated at constant luma).
        - "Lb": Apply the operation to the luma, with highlights protection by blending
          (after the operation, the out-of-range pixels are blended with channels = "RGB").

      trans(bool, optional): If True (default), embed the transormation in the output image as
        output.trans (see Image.apply_channels).

    Returns:
      Image: The stretched image.
    """
    if gamma <= 0.: raise ValueError("Error, gamma must be > 0.")
    return self.apply_channels(lambda channel: stf.gamma_stretch_function(channel, gamma), channels, trans = trans)

  def adjust_midtone_levels(self, midtone, shadow = 0., highlight = 1., low = 0., high = 1., channels = "", trans = True):
    """Adjust shadow, midtone, highlight, low and high levels in selected channels of the image.

    This method:
      1) Clips the selected channels in the [shadow, highlight] range and maps [shadow, highlight] to [0, 1].
      2) Applies the midtone stretch function f(x) = (m-1)*x/((2*m-1)*x-m),
         with m = (midtone-shadow)/(highlight-shadow) the remapped midtone.
      3) Maps [low, high] to [0, 1] and clips the output data in the [0, 1] range.

    Args:
      midtone (float): The input midtone level (expected in ]0, 1[).
      shadow (float, optional): The input shadow level (default 0; expected < midtone).
      highlight (float, optional): The input highlight level (default 1; expected > midtone).
      low (float, optional): The "low" output level (default 0; expected <= 0).
      high (float, optional): The "high" output level (default 1; expected >= 1).
      channels (str, optional): The selected channels:

        - An empty string (default): Apply the operation to all channels (RGB, HSV and grayscale images).
        - A combination of "1", "2", "3" (or equivalently "R", "G", "B" for RGB images): Apply the
          operation to the first/second/third channel (RGB, HSV and grayscale images).
        - "V": Apply the operation to the HSV value (RGB, HSV and and grayscale images).
        - "S": Apply the operation to the HSV saturation (RGB and HSV images).
        - "L": Apply the operation to the luma (RGB and grayscale images).
        - "Ls": Apply the operation to the luma, with highlights protection by desaturation
          (after the operation, the out-of-range pixels are desaturated at constant luma).
        - "Lb": Apply the operation to the luma, with highlights protection by blending
          (after the operation, the out-of-range pixels are blended with channels = "RGB").

      trans(bool, optional): If True (default), embed the transormation in the output image as
        output.trans (see Image.apply_channels).

    Returns:
      Image: The stretched image.
    """
    if midtone < .0001 or midtone > .9999: raise ValueError("Error, midtone must be >= 0.0001 and <= 0.9999.")
    if midtone-shadow < .0001: raise ValueError("Error, midtone-shadow must be >= 0.0001.")
    if highlight-midtone < .0001: raise ValueError("Error, highlight-midtone must be >= 0.0001.")
    if low > 0.:
      low = 0.
      print("Warning, changed low = 0.")
    if high < 1.:
      high = 1.
      print("Warning, changed high = 1.")
    output = self.apply_channels(lambda channel: stf.midtone_levels_stretch_function(channel, shadow, midtone, highlight, low, high), channels, trans = trans)
    if trans: output.trans.xticks = [shadow, midtone, highlight]
    return output



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.0.0 / 2024.12.15
# Sphinx OK.

"""Plotly backend for Jupyter-lab interface."""

import numpy as np
import plotly.express as px
import plotly.graph_objects as go
import plotly.io as pio
from plotly.subplots import make_subplots
pio.renderers.default = "jupyterlab"

from . import params
from .utils import prepare_images

from ..equimage import Image

#####################
# Helper functions. #
#####################

def _figure_prepared_image_(image, width = -1, hover = False, template = "plotly_dark"):
  """Prepare a ploty figure for the input (prepared) image.

  Args:
    image (numpy.ndarray): The prepared image (processed by utils.prepare_images).3
    width (int, optional): The width of the figure (defaults to `jupyter.params.maxwidth` if negative).
    hover (bool, optional): If True, show the image data on hover (default False).
      Warning: Setting hover = True can slow down display a lot !
    template (str, optional): The template for the figure (default "plotly_dark").

  Returns:
    plotly.graph_objects.Figure: A plotly figure with the image.
  """
  if width <= 0: width = params.maxwidth
  # Plot image.
  figure = px.imshow(image, zmin = 0., zmax = 1., aspect = "equal", binary_string = not hover)
  figure.update_traces(name = "", hovertemplate = "(%{x}, %{y}): %{z}" if hover else "(%{x}, %{y})")
  layout = go.Layout(template = template,
                     width = width+params.lmargin+params.rmargin, height = width*image.shape[0]/image.shape[1]+params.bmargin+params.tmargin,
                     margin = go.layout.Margin(l = params.lmargin, r = params.rmargin, b = params.bmargin, t = params.tmargin, autoexpand = True))
  figure.update_layout(layout)
  return figure

def _figure_image_(image, sampling = -1, width = -1, hover = False, template = "plotly_dark"):
  """Prepare a ploty figure for the input image.

  Args:
    image: The image (Image object or numpy.ndarray).
    sampling (int, optional): Downsampling rate (defaults to `jupyter.params.sampling` if negative).
      Only image[::sampling, ::sampling] is shown, to speed up display.
    width (int, optional): The width of the figure (defaults to `jupyter.params.maxwidth` if negative).
    hover (bool, optional): If True, show the image data on hover (default False).
      Warning: Setting hover = True can slow down display a lot !
    template (str, optional): The template for the figure (default "plotly_dark").

  Returns:
    plotly.graph_objects.Figure: A plotly figure with the image.
  """
  return _figure_prepared_image_(prepare_images(image, sampling = sampling), width = width, hover = hover, template = template)

def _figure_histograms_(image, channels = "", log = True, width = -1, xlabel = "Level", trans = None, template = "plotly_dark"):
  """Prepare a plotly figure with the histograms of an image.

  Args:
    image (Image): The image.
    channels (str, optional): The channels of the histograms (default "" = "RGBL" for red, green, blue, luma).
    log (bool, optional): If True (default), plot the histogram counts in log scale.
    width (int, optional): The width of the figure (defaults to `jupyter.params.maxwidth` if negative).
    xlabel (str, optional): The x axis label of the plot (default "Level").
    trans (optional): A container with an histogram transformation (see Image.apply_channels), plotted on top
      of the histograms (default None).
    template (str, optional): The template for the figure (default "plotly_dark").

  Returns:
    plotly.graph_objects.Figure: A plotly figure with the histograms of the image.
  """
  if width <= 0: width = params.maxwidth
  # Prepare histograms.
  if not issubclass(type(image), Image):
    print("The histograms can only be displayed for Image objects.")
    return None
  if channels == "":
    hists = getattr(image, "hists", None)
    if hists is None: hists = image.histograms()
  else:
    hists = image.histograms(channels = channels)
  # Set-up lines.
  mline = dict(color = params.mlinecolor)
  mline2 = dict(color = params.mlinecolor, width = 2)
  mlinedot1 = dict(color = params.mlinecolor, dash = "dot", width = 1)
  mlinedash1 = dict(color = params.mlinecolor, dash = "dash", width = 1)
  mlinedashdot1 = dict(color = params.mlinecolor, dash = "dashdot", width = 1)
  # Plot histograms.
  figure = make_subplots(specs = [[dict(secondary_y = trans is not None, r = -0.06)]])
  updatemenus = []
  ntraces = 0
  for channel in hists.values():
    ntraces += 1
    midpoints = (channel.edges[1:]+channel.edges[:-1])/2.
    figure.add_trace(go.Scatter(x = midpoints, y = channel.counts, name = channel.name, mode = "lines", line = dict(color = channel.color, width = 2)),
                     secondary_y = False)
  figure.add_vline(x = 0., line = mlinedashdot1, secondary_y = False)
  figure.add_vline(x = 1., line = mlinedashdot1, secondary_y = False)
  figure.update_xaxes(title_text = xlabel, ticks = "inside", rangemode = "tozero")
  figure.update_yaxes(title_text = "Count", ticks = "inside", rangemode = "tozero", secondary_y = False)
  if log:
    figure.update_yaxes(type = "log", secondary_y = False)
    active =  0
  else:
    active = -1
  # Add lin/log toggle button.
  ybutton = 1.+.025*1024./width
  buttons = [dict(label = "lin/log", method = "relayout", args = [{"yaxis.type": "log"}], args2 = [{"yaxis.type": "linear"}])]
  updatemenus.append(dict(type = "buttons", buttons = buttons, active = active, showactive = False,
                          xanchor = "left", x = 0., yanchor = "bottom", y = ybutton))
  # Plot transformation.
  if trans is not None:
    if trans.type == "hist":
      cef = np.log(np.maximum(np.gradient(trans.y, trans.x), 1.e-12))
      figure.add_trace(go.Scatter(x = trans.x, y = trans.y, name = trans.ylabel, mode = "lines", line = mline2, showlegend = False),
                       secondary_y = True)
      figure.add_trace(go.Scatter(x = trans.x, y = cef, name = f"log {trans.ylabel}'", mode = "lines", line = mline2, showlegend = False, visible = False),
                       secondary_y = True)
      figure.add_trace(go.Scatter(x = [0., 1.], y = [0., 0.], name = "", mode = "lines", line = mlinedashdot1, showlegend = False),
                       secondary_y = True)
      figure.add_trace(go.Scatter(x = [0., 1.], y = [1., 1.], name = "", mode = "lines", line = mlinedashdot1, showlegend = False),
                       secondary_y = True)
      figure.add_trace(go.Scatter(x = [0., 1.], y = [0., 1.], name = "", mode = "lines", line = mlinedot1, showlegend = False),
                       secondary_y = True)
      xticks = getattr(trans, "xticks", None)
      if xticks is not None:
        for xtick in xticks:
          figure.add_vline(x = xtick, line = mlinedash1, secondary_y = True)
      ftitle = f"{trans.ylabel}({trans.channels})"
      ceftitle = f"log {trans.ylabel}'({trans.channels})"
      figure.update_yaxes(title_text = ftitle, titlefont = mline, ticks = "inside", tickfont = mline, showgrid = False, rangemode = "tozero", secondary_y = True)
      # Add f/log f' toggle button.
      keepvisible = ntraces*[True]
      buttons = [dict(label = "f/log f'", method = "update",
                      args  = [{"visible": keepvisible+[False, True, True, False, False]}, {"yaxis2.title": ceftitle}],
                      args2 = [{"visible": keepvisible+[True, False, True, True, True]}, {"yaxis2.title": ftitle}])]
      xbutton = .066*1024./width
      updatemenus.append(dict(type = "buttons", buttons = buttons, active = -1, showactive = False,
                              xanchor = "left", x = xbutton, yanchor = "bottom", y = ybutton))
    elif trans.type == "hue":
      figure.add_trace(go.Scatter(x = trans.x, y = trans.y, name = trans.ylabel, mode = "lines", line = mline2, showlegend = False),
                       secondary_y = True)
      figure.add_trace(go.Scatter(x = trans.xm, y = trans.ym, name = trans.ylabel, mode = "markers", marker = dict(size = 16, color = trans.cm), showlegend = False),
                       secondary_y = True)
      figure.add_trace(go.Scatter(x = [0., 1.], y = [0., 0.], name = "", mode = "lines", line = mlinedashdot1, showlegend = False),
                       secondary_y = True)
      figure.update_xaxes(range = [0., 1.])
      figure.update_yaxes(title_text = trans.ylabel, titlefont = mline, ticks = "inside", tickfont = mline, showgrid = False, rangemode = "tozero", secondary_y = True)
    else:
      print(f"Can not handle transformations of type '{trans.type}'.")
  # Finalize layout.
  xlegend = 1.+.05*1024./width
  layout = go.Layout(template = template,
                     width = width+params.lmargin+params.rmargin, height = width/3+params.bmargin+params.tmargin,
                     margin = go.layout.Margin(l = params.lmargin, r = params.rmargin, b = params.bmargin, t = params.tmargin, autoexpand = True),
                     legend = dict(xanchor = "left", x = xlegend, yanchor = "top", y = 1.))
  figure.update_layout(layout, updatemenus = updatemenus)
  return figure

def _figure_statistics_(image, channels = "", width = -1, rowheight = -1, template = "plotly_dark"):
  """Prepare a plotly table with the statistics of an image.

  Args:
    image (Image): The image.
    channels (str, optional): The channels of the statistics (default "" = "RGBL" for red, green, blue, luma).
    width (int, optional): The width of the table (defaults to `jupyter.params.maxwidth` if negative).
    rowheight (int, optional): The height of the rows (default to jupyter.params.rowheight if negative).
    template (str, optional): The template for the figure (default "plotly_dark").

  Returns:
    plotly.graph_objects.Figure: A plotly figure with the table of the statistics of the image.
  """
  if width <= 0: width = params.maxwidth
  if rowheight <= 0: rowheight = params.rowheight
  # Prepare statistics.
  if not issubclass(type(image), Image):
    print("The statistics can only be displayed for Image objects.")
    return None
  if channels == "":
    stats = getattr(image, "stats", None)
    if stats is None: stats = image.statistics()
  else:
    stats = image.statistics(channels = channels)
  # Create table.
  columns = [[], [], [], [], [], [], [], []]
  for channel in stats.values():
    columns[0].append(channel.name)
    columns[1].append(f"{channel.minimum:.5f}")
    if channel.percentiles is not None:
      columns[2].append(f"{channel.percentiles[0]:.5f}")
      columns[3].append(f"{channel.percentiles[1]:.5f}")
      columns[4].append(f"{channel.percentiles[2]:.5f}")
    else:
      columns[2].append("None")
      columns[3].append("None")
      columns[4].append("None")
    columns[5].append(f"{channel.maximum:.5f}")
    columns[6].append(f"{channel.zerocount} ({100.*channel.zerocount/channel.npixels:.2f}%)")
    columns[7].append(f"{channel.outcount} ({100.*channel.outcount/channel.npixels:.2f}%)")
  align = ["left"]+7*["right"]
  header = dict(values = ["Channel", "Minimum", "25%", "50%", "75%", "Maximum", "Shadowed", "Highlighted"], align = align, height = rowheight)
  cells = dict(values = columns, align = align, height = rowheight)
  table = go.Table(header = header, cells = cells, columnwidth = [1, 1, 1, 1, 1, 1, 1.5, 1.5])
  # Create figure.
  figure = go.Figure(data = table)
  layout = go.Layout(template = template,
                     width = width+params.lmargin+params.rmargin, height = (len(stats)+1)*rowheight+params.bmargin+params.tmargin,
                     margin = go.layout.Margin(l = params.lmargin, r = params.rmargin, b = 32, t = 32, autoexpand = True))
  figure.update_layout(layout)
  return figure

#####################
# Plotly interface. #
#####################

def show(image, histograms = False, statistics = False, sampling = -1, width = -1, hover = False, renderer = None):
  """Show an image using plotly.

  Args:
    image: The image (Image object or numpy.ndarray; must be an Image object if histograms or statistics is True).
    histograms (optional): If True or a string, show the histograms of the image. The string lists the
      channels of the histograms (e.g. "RGBL" for red, green, blue, luma). Default is False.
    statistics (optional): If True or a string, show the statistics of the image. The string lists the
      channels of the statistics (e.g. "RGBL" for red, green, blue, luma). Default is False.
    sampling (int, optional): Downsampling rate (defaults to `jupyter.params.sampling` if negative).
      Only image[::sampling, ::sampling] is shown, to speed up display.
    width (int, optional): The width of the figure (defaults to `jupyter.params.maxwidth` if negative).
    hover (bool, optional): If True, show the image data on hover (default False).
      Warning: Setting hover = True can slow down display a lot !
    renderer (str, optional): The plotly renderer (default None = "jupyterlab").
  """
  figure = _figure_image_(image, sampling = sampling, width = width, hover = hover)
  figure.show(renderer)
  if histograms is not False:
    if histograms is True: histograms = ""
    show_histograms(image, channels = histograms, width = width, renderer = renderer)
  if statistics is not False:
    if statistics is True: statistics = ""
    show_statistics(image, channels = statistics, width = width, renderer = renderer)

def show_histograms(image, channels = "", log = True, width = -1, xlabel = "Level", trans = None, renderer = None):
  """Plot the histograms of an image using plotly.

  Args:
    image (Image): The image.
    channels (str, optional): The channels of the histograms (default "" = "RGBL" for red, green, blue, luma).
    log (bool, optional): If True (default), plot the histogram counts in log scale.
    width (int, optional): The width of the figure (defaults to `jupyter.params.maxwidth` if negative).
    xlabel (str, optional): The x axis label of the plot (default "Level").
    trans (optional): A container with an histogram transformation (see Image.apply_channels), plotted on top
      of the histograms (default None).
    renderer (str, optional): The plotly renderer (default None = "jupyterlab").
  """
  figure = _figure_histograms_(image, channels = channels, log = log, width = width, xlabel = xlabel, trans = trans)
  if figure is not None: figure.show(renderer)

def show_statistics(image, channels = "", width = -1, rowheight = -1, renderer = None):
  """Display a table with the statistics of an image using plotly.

  Args:
    image (Image): The image.
    channels (str, optional): The channels of the statistics (default "" = "RGBL" for red, green, blue, luma).
    width (int, optional): The width of the table (defaults to `jupyter.params.maxwidth` if negative).
    rowheight (int, optional): The height of the rows (default to `jupyter.params.rowheight` if negative).
    renderer (str, optional): The plotly renderer (default None = "jupyterlab").
  """
  figure = _figure_statistics_(image, channels = channels, width = width, rowheight = rowheight)
  if figure is not None: figure.show(renderer)

def show_t(image, channels = "RGBL", sampling = -1, width = -1, hover = False, renderer = None):
  """Show an image embedding an histogram transformation using plotly.

  Displays the input histograms with the transformation curve, the output histograms, and the output image.

  Args:
    image (Image): The output image (must embed a transformation image.trans - see Image.apply_channels).
    channels (str, optional): The channels of the histograms (default "RGBL" for red, green, blue, luma).
      The channels of the transformation are added if needed.
    sampling (int, optional): Downsampling rate (defaults to `jupyter.params.sampling` if negative).
      Only image[::sampling, ::sampling] is shown, to speed up display.
    width (int, optional): The width of the figure (defaults to `jupyter.params.maxwidth` if negative).
    hover (bool, optional): If True, show the image data on hover (default False).
      Warning: Setting hover = True can slow down display a lot !
    renderer (str, optional): The plotly renderer (default None = "jupyterlab").
  """
  if not issubclass(type(image), Image): print("The transformations can only be displayed for Image objects.")
  trans = getattr(image, "trans", None)
  if trans is None:
    print("There is no transformation embedded in the input image.")
    return
  reference = trans.input
  if trans.type == "hist":
    for c in trans.channels:
      if c in "RGBVSL" and not c in channels:
        channels += c
  show_histograms(reference, channels = channels, log = True, width = width, xlabel = "Input level", trans = trans, renderer = renderer)
  show_histograms(image, channels = channels, log = True, width = width, xlabel = "Output level", renderer = renderer)
  show(image, histograms = False, statistics = False, sampling = sampling, width = width, hover = hover, renderer = renderer)

def light_curve(image, reference, maxpoints = 32768, width = -1, renderer = None):
  """Plot light curve (scatter plot of an output image vs an input reference channel).

  Args:
    image (numpy.ndarray): The output image channel (luma, ...) as an array with shape (height, width).
    reference (numpy.ndarray): The input reference channel as an array with shape (height, width).
    maxpoints (int, optional): The maximum number of points in the scatter plot. The image and reference will be
      sampled accordingly (default 32768).
    width (int, optional): The width of the figure (defaults to `jupyter.params.maxwidth` if negative).
    renderer (str, optional): The plotly renderer (default None = "jupyterlab").
  """
  if width <= 0: width = params.maxwidth
  # Prepare light curve.
  if image.ndim != 2 or reference.ndim != 2:
    raise ValueError("Error, image and reference must be 2D arrays.")
  if image.shape != reference.shape:
    raise ValueError("Error, image and reference must have the same shape.")
  input = reference.ravel()
  output = image.ravel()
  step = int(np.ceil(len(input)/maxpoints))
  # Set-up lines.
  mlinedot1 = dict(color = params.mlinecolor, dash = "dot", width = 1)
  mlinedashdot1 = dict(color = params.mlinecolor, dash = "dashdot", width = 1)
  # Plot light curve.
  figure = go.Figure()
  figure.add_trace(go.Scatter(x = input[::step], y = output[::step],
                              name = "Light curve", mode = "markers", marker = dict(size = 4, color = "lightslategray"), showlegend = False))
  figure.add_trace(go.Scatter(x = [0., 1.], y = [0., 0.], name = "", mode = "lines", line = mlinedashdot1, showlegend = False))
  figure.add_trace(go.Scatter(x = [0., 1.], y = [1., 1.], name = "", mode = "lines", line = mlinedashdot1, showlegend = False))
  figure.add_trace(go.Scatter(x = [0., 0.], y = [0., 1.], name = "", mode = "lines", line = mlinedashdot1, showlegend = False))
  figure.add_trace(go.Scatter(x = [1., 1.], y = [0., 1.], name = "", mode = "lines", line = mlinedashdot1, showlegend = False))
  figure.add_trace(go.Scatter(x = [0., 1.], y = [0., 1.], name = "", mode = "lines", line = mlinedot1, showlegend = False))
  figure.update_xaxes(title_text = "Input level", ticks = "inside", rangemode = "tozero")
  figure.update_yaxes(title_text = "Output level", ticks = "inside", rangemode = "tozero")
  layout = go.Layout(template = "plotly_dark",
                     width = width+params.lmargin+params.rmargin, height = width+params.bmargin+params.tmargin,
                     margin = go.layout.Margin(l = params.lmargin, r = params.rmargin, b = params.bmargin, t = params.tmargin, autoexpand = True))
  figure.update_layout(layout)
  figure.show(renderer)



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.0.0 / 2024.12.15
# Sphinx OK.

"""Image masks."""

import numpy as np
import scipy.ndimage as ndimg
import skimage.morphology as skimo

from . import params

def threshold_mask(filtered, threshold, extend = 0, smooth = 0):
  """Set-up a threshold mask.

  Returns the pixels such that filtered >= threshold.

  See also:
    Image.filter

  Args:
    filtered (numpy.ndarray): The output of a filter (local average, ...) applied to an image (see Image.filter).
    threshold (float): The threshold for the mask. The mask is 1 wherever filtered >= threshold, and 0 elsewhere.
    extend (int, optional): Once computed, the mask can be extended by extend pixels (default 0).
    smooth (int, optional): Once extended, the edges of the mask can be smoothed over smooth pixels (default 0).

  Returns:
    numpy.ndarray: The mask as an array with the same shape as filtered (*not* converted to a grayscale Image object).
  """
  # Threshold the filter.
  mask = (filtered >= threshold)
  # Extend the mask.
  if extend > 0: mask = skimo.isotropic_dilation(mask, extend)
  # Smooth the mask.
  mask = mask.astype(params.IMGTYPE)
  if smooth > 0:
    kernel = skimo.disk(smooth, dtype = params.IMGTYPE)
    kernel /= np.sum(kernel)
    mask = ndimg.convolve(mask, kernel, mode = "reflect")
  return mask

#####################################
# For inclusion in the Image class. #
#####################################

class Mixin:
  """To be included in the Image class."""

  def filter(self, channel, filter, radius, mode = "reflect"):
    """Apply a spatial filter to a selected channel of the image.

    The main purpose of this method is to prepare masks for image processing.

    See also:
      threshold_mask

    Args:
      channel (str): The selected channel:

        - "1", "2", "3" (or equivalently "R", "G", "B" for RGB images):
          Apply the filter to the first/second/third channel (RGB, HSV and grayscale images).
        - "V": Apply the filter to the HSV value (RGB, HSV and and grayscale images).
        - "L": Apply the filter to the luma (RGB and grayscale images).
        - "L*": Apply the filter to the CIE lightness (RGB and grayscale images).
          Warning: The CIE lightness is normalized within [0, 1] instead of [0, 100] !

      filter (str): The filter:

        - "mean": Return the average of the channel within a disk around each pixel.
        - "median": Return the median of the channel within a disk around each pixel.
        - "gaussian": Return the gaussian average of the channel around each pixel.
        - "maximum": Return the maximum of the channel within a disk around each pixel.

      radius (float): The radius of the disk in pixels. The standard deviation for gaussian average is radius/3.
      mode (str, optional): How to extend the image across its boundaries:

        - "reflect" (default): the image is reflected about the edge of the last pixel (abcd -> dcba|abcd|dcba).
        - "mirror": the image is reflected about the center of the last pixel (abcd -> dcb|abcd|cba).
        - "nearest": the image is padded with the value of the last pixel (abcd -> aaaa|abcd|dddd).
        - "zero": the image is padded with zeros (abcd -> 0000|abcd|0000).

    Returns:
      numpy.ndarray: A (image height, image width) array with the output of the filter,
      *not* converted to a grayscale Image object.
    """
    if mode == "zero": mode = "constant" # Translate modes.
    if channel == "1":
      data = self.image[0]
    elif channel == "2":
      self.check_is_not_gray()
      data = self.image[1]
    elif channel == "3":
      self.check_is_not_gray()
      data = self.image[2]
    elif channel == "R":
      self.check_color_model("RGB")
      data = self.image[0]
    elif channel == "G":
      self.check_color_model("RGB")
      data = self.image[1]
    elif channel == "B":
      self.check_color_model("RGB")
      data = self.image[2]
    elif channel == "V":
      data = self.value()
    elif channel == "L":
      data = self.luma()
    elif channel == "L*":
      data = self.lightness()/100.
    else:
      raise ValueError(f"Error, unknown or incompatible channel '{channel}'.")
    if filter == "gaussian":
      return ndimg.gaussian_filter(data, sigma = radius/3., mode = mode, cval = 0.)
    elif filter == "mean":
      kernel = skimo.disk(radius, dtype = params.IMGTYPE)
      kernel /= np.sum(kernel)
      return ndimg.convolve(data, kernel, mode = mode, cval = 0.)
    elif filter == "median":
      return ndimg.median_filter(data, footprint = skimo.disk(radius), mode = mode, cval = 0.)
    elif filter == "maximum":
      return ndimg.maximum_filter(data, footprint = skimo.disk(radius), mode = mode, cval = 0.)
    else:
      raise ValueError(f"Error, unknown filter '{filter}'.")



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.0.0 / 2024.12.15
# Sphinx OK.

"""Image processing parameters."""

import numpy as np

# Data type used for images (either np.float32 or np.float64).

IMGTYPE = np.float32

# Expected accuracy in np.float32/np.float64 calculations.

IMGTOL = 1.e-6 if IMGTYPE is np.float32 else 1.e-9

# Use imageio or imread module ?

IMAGEIO = False

# Exclude pixels <= 0 or >= 1 from the percentiles in image statistics ?

exclude01 = False

# Number of x mesh points for transformation function plots y = f(x in [0, 1]).

ntranslo = 128 # Low  resolution.
ntranshi = 256 # High resolution.

# Number of bins in the histograms.

maxhistbins = 8192 # Maximum number of bins within [0, 1].
defhistbins = 0    # Default number of bins within [0, 1] (see set_default_hist_bins).

def set_max_hist_bins(n):
  """Set the maximum number of bins in the histograms.

  Args:
    n (int): The maximum number of bins within [0, 1].
  """
  global maxhistbins
  maxhistbins = n

def set_default_hist_bins(n):
  """Set the default number of bins in the histograms.

  Args:
    n (int): If strictly positive, the default number of bins within [0, 1].
      (practically limited to `equimage.params.maxhistbins`). If zero, the number
      of bins is computed according to the statistics of each image. If strictly
      negative, the number of bins is set to `equimage.params.maxhistbins`.
  """
  global defhistbins
  defhistbins = n if n >= 0 else maxhistbins

# Weights of the RGB components in the luma.

rgbluma = IMGTYPE((.299, .587, .114))

def get_RGB_luma():
  """Return the RGB weights rgbluma of the luma.

  The luma L of an image is the average of the RGB components weighted by rgbluma:

    L = rgbluma[0]*image[0]+rgbluma[1]*image[1]+rgbluma[2]*image[2]

  Returns:
    tuple: The (red, blue, green) weights rgbluma of the luma.
  """
  return rgbluma

def set_RGB_luma(rgb):
  """Set the RGB weights of the luma.

  Args:
    rgb: The RGB weights of the luma as:

      - a tuple, list or array of the (red, green, blue) weights. They will be normalized so that their sum is 1.
      - the string "uniform": the RGB weights are set to (1/3, 1/3, 1/3).
      - the string "human": the RGB weights are set to (.299, .587, .114). The luma is then the luminance for lRGB images,
        and an approximate substitute for the lightness for sRGB images.
  """
  if isinstance(rgb, str):
    if rgb == "uniform":
      set_RGB_luma((1./3., 1./3., 1./3.))
    elif rgb == "human":
      set_RGB_luma((.299, .587, .114))
    else:
      raise ValueError("Error, the input rgb weights must be an array with three scalar elements, the string 'uniform' or the string 'human'.")
  else:
    w = np.array(rgb, dtype = IMGTYPE)
    if w.shape != (3,): raise ValueError("Error, the input rgb weights must be an array with three scalar elements, the string 'uniform' or the string 'human'.")
    if any(w < 0.): raise ValueError("Error, the input rgb weights must be >= 0.")
    s = np.sum(w)
    if s == 0.: raise ValueError("Error, the sum of the input rgb weights must be > 0.")
    global rgbluma ; rgbluma = w/s
    print(f"Luma = {rgbluma[0]:.3f}R+{rgbluma[1]:.3f}G+{rgbluma[2]:.3f}B.")



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.0.0 / 2024.12.15
# Sphinx OK.

"""Interface with scikit-image."""

import numpy as np
import skimage as skim

#####################################
# For inclusion in the Image class. #
#####################################

class Mixin:
  """To be included in the Image class."""

  ###########
  # Filter. #
  ###########

  def gaussian_filter(self, sigma, mode = "reflect", channels = ""):
    """Convolve (blur) selected channels of the image with a gaussian.

    Args:
      sigma (float): The standard deviation of the gaussian (pixels).
      mode (str, optional): How to extend the image across its boundaries:

        - "reflect" (default): the image is reflected about the edge of the last pixel (abcd -> dcba|abcd|dcba).
        - "mirror": the image is reflected about the center of the last pixel (abcd -> dcb|abcd|cba).
        - "nearest": the image is padded with the value of the last pixel (abcd -> aaaa|abcd|dddd).
        - "zero": the image is padded with zeros (abcd -> 0000|abcd|0000).

      channels (str, optional): The selected channels:

        - An empty string (default): Apply the operation to all channels (RGB, HSV and grayscale images).
        - A combination of "1", "2", "3" (or equivalently "R", "G", "B" for RGB images):
          Apply the operation to the first/second/third channel (RGB, HSV and grayscale images).
        - "V": Apply the operation to the HSV value (RGB, HSV and and grayscale images).
        - "S": Apply the operation to the HSV saturation (RGB and HSV images).
        - "L": Apply the operation to the luma (RGB and grayscale images).
        - "Ls": Apply the operation to the luma, with highlights protection by desaturation.
          (after the operation, the out-of-range pixels are desaturated at constant luma).
        - "Lb": Apply the operation to the luma, with highlights protection by blending.
          (after the operation, the out-of-range pixels are blended with channels = "RGB").

    Returns:
      Image: The processed image.

    See also:
      skimage.filters.gaussian
    """
    if mode == "zero": mode = "constant" # Translate modes.
    return self.apply_channels(lambda channel: skim.filters.gaussian(channel, channel_axis = 0, sigma = sigma, mode = mode, cval = 0.), channels)

  def butterworth_filter(self, cutoff, order = 2, padding = 0, channels = ""):
    """Apply a Butterworth low-pass filter to selected channels of the image.

    The Butterworh filter reads in the frequency domain:

    .. math::
      H(f) = 1/(1+(f/f_c)^{2n})

    where n is the order of the filter and fc the cut-off frequency.
    The data are Fast-Fourier Transformed back and forth to apply the filter.

    Args:
      cutoff (float): The normalized cutoff frequency in [0, 1] (fc = (1-cutoff)fs/2 with fs the FFT sampling frequency).
      order (int, optional): The order of the filter (default 2).
      padding (int, optional): Number of pixels to pad the image with (default 0; increase if the filter leaves visible
        artifacts on the edges).
      channels (str, optional): The selected channels:

        - An empty string (default): Apply the operation to all channels (RGB, HSV and grayscale images).
        - A combination of "1", "2", "3" (or equivalently "R", "G", "B" for RGB images):
          Apply the operation to the first/second/third channel (RGB, HSV and grayscale images).
        - "V": Apply the operation to the HSV value (RGB, HSV and and grayscale images).
        - "S": Apply the operation to the HSV saturation (RGB and HSV images).
        - "L": Apply the operation to the luma (RGB and grayscale images).
        - "Ls": Apply the operation to the luma, with highlights protection by desaturation.
          (after the operation, the out-of-range pixels are desaturated at constant luma).
        - "Lb": Apply the operation to the luma, with highlights protection by blending.
          (after the operation, the out-of-range pixels are blended with channels = "RGB").

    Returns:
      Image: The processed image.

    See also:
      skimage.filters.butterworth
    """
    return self.apply_channels(lambda channel: skim.filters.butterworth(channel, channel_axis = 0, cutoff_frequency_ratio = (1.-cutoff)/2.,
                               order = order, npad = padding, squared_butterworth = True), channels)

  def unsharp_mask(sigma, strength, channels = ""):
    """Apply an unsharp mask to selected channels of the image.

    Given a channel Cin, returns

      Cout = Cin + strength*(Cin-BLUR(Cin))

    where BLUR(Cin) is the convolution of Cin with a gaussian of standard deviation sigma.
    As BLUR(Cin) is a low-pass filter, Cin-BLUR(Cin) is a high-pass filter whose output is
    admixed in the original image. This enhances details; the larger the mixing strength,
    the sharper the image, at the expense of noise and fidelity.

    Args:
      sigma (float): The standard deviation of the gaussian (pixels).
      strength (float): The mixing strength.
      channels (str, optional): The selected channels:

        - An empty string (default): Apply the operation to all channels (RGB, HSV and grayscale images).
        - A combination of "1", "2", "3" (or equivalently "R", "G", "B" for RGB images):
          Apply the operation to the first/second/third channel (RGB, HSV and grayscale images).
        - "V": Apply the operation to the HSV value (RGB, HSV and and grayscale images).
        - "S": Apply the operation to the HSV saturation (RGB and HSV images).
        - "L": Apply the operation to the luma (RGB and grayscale images).
        - "Ls": Apply the operation to the luma, with highlights protection by desaturation.
          (after the operation, the out-of-range pixels are desaturated at constant luma).
        - "Lb": Apply the operation to the luma, with highlights protection by blending.
          (after the operation, the out-of-range pixels are blended with channels = "RGB").

    Returns:
      Image: The processed image.

    See also:
      skimage.filters.unsharp_mask
    """
    return self.apply_channels(lambda channel: skim.filters.unsharp_mask(channel, channel_axis = 0, radius = radius, amount = strength), channels)

  ############
  # Denoise. #
  ############

  def estimate_noise(self):
    """Estimate the rms noise of the image, averaged over all channels.

    Returns:
      float: The rms noise of the image, averaged over the channels.

    See also:
      skimage.restoration.estimate_sigma

    To do:
      Estimate the noise in arbitrary channels.
    """
    return skim.restoration.estimate_sigma(self.image, channel_axis = 0, average_sigmas = True)

  def wavelets_filter(self, sigma, wavelet = "coif4", mode = "soft", method = "BayesShrink", shifts = 0, channels = "L"):
    """Wavelets filter for denoising selected channels of the image.

    Performs a wavelets transform on the selected channels and filters the wavelets to reduce noise.

    Args:
      sigma: The estimated noise standard deviation used to compute the wavelets filter threshold.
        The larger sigma, the smoother the output image.
      wavelet (str, optional): The wavelets used to decompose the image (default "coif4").
        Can be any of the orthogonal wavelets of `pywt.wavelist`. Recommended wavelets are:

          - Daubechies wavelets ("db1"..."db8"),
          - Symlets ("sym2"..."sym8"),
          - Coiflets ("coif1"..."coif8").

      mode (str, optional): Denoising method [either "soft" (default) or "hard"].
      method (str, optional): Thresholding method [either "BayesShrink" (default) or "VisuShrink"].
        Separate thresholds are applied to the wavelets bands for "BayesShrink", whereas a
        single threshold is applied for "VisuShrink" (best in principle for Gaussian noise,
        but may appear overly smooth).
      shifts (int, optional): Number of spin cycles (default 0). The wavelets transform is not
        shift-invariant. To mimic a shift-invariant transform as best as possible, the output image
        is an average of the original image shifted shifts times in each direction, filtered, then
        shifted back to the original position.
      channels (str, optional): The selected channels:

        - An empty string (default): Apply the operation to all channels (RGB, HSV and grayscale images).
        - A combination of "1", "2", "3" (or equivalently "R", "G", "B" for RGB images):
          Apply the operation to the first/second/third channel (RGB, HSV and grayscale images).
        - "V": Apply the operation to the HSV value (RGB, HSV and and grayscale images).
        - "S": Apply the operation to the HSV saturation (RGB and HSV images).
        - "L": Apply the operation to the luma (RGB and grayscale images).
        - "Ls": Apply the operation to the luma, with highlights protection by desaturation.
          (after the operation, the out-of-range pixels are desaturated at constant luma).
        - "Lb": Apply the operation to the luma, with highlights protection by blending.
          (after the operation, the out-of-range pixels are blended with channels = "RGB").

    Returns:
      Image: The processed image.

    See also:
      skimage.restoration.denoise_wavelet,
      skimage.restoration.cycle_spin
    """
    kwargs = dict(channel_axis = -1, sigma = sigma, wavelet = wavelet, wavelet_levels = None, mode = mode,
                  method = method, convert2ycbcr = False, rescale_sigma = True)
    return self.apply_channels(lambda channel: skim.restoration.cycle_spin(channel, channel_axis = 0, max_shifts = shifts,
                               func = skim.restoration.denoise_wavelet, func_kw = kwargs, num_workers = None), channels)

  def bilateral_filter(self, sigma_space, sigma_color = .1, mode = "reflect", channels = ""):
    r"""Bilateral filter for denoising selected channels of the image.

    The bilateral filter convolves the selected channel(s) Cin with a gaussian gs of standard deviation
    sigma_space weighted by a gaussian gc in color space (with standard deviation sigma_color):

    .. math::
      C_{out}(\mathbf{r}) \sim \sum_{\mathbf{r}'} C_{in}(\mathbf{r}') g_s(|\mathbf{r}-\mathbf{r}'|) g_c(|C_{in}(\mathbf{r})-C_{in}(\mathbf{r}')|)

    Therefore, the bilateral filter averages the neighboring pixels whose colors are sufficiently similar.
    The bilateral filter may tend to produce cartoon-like (piecewise-constant) images.

    Args:
      sigma_space (float): The standard deviation of the gaussian in real space (pixels).
      sigma_color (float, optional): The standard deviation of the gaussian in color space (default 0.1).
      mode (str, optional): How to extend the image across its boundaries:

        - "reflect" (default): the image is reflected about the edge of the last pixel (abcd -> dcba|abcd|dcba).
        - "mirror": the image is reflected about the center of the last pixel (abcd -> dcb|abcd|cba).
        - "nearest": the image is padded with the value of the last pixel (abcd -> aaaa|abcd|dddd).
        - "zero": the image is padded with zeros (abcd -> 0000|abcd|0000).

      channels (str, optional): The selected channels:

        - An empty string (default): Apply the operation to all channels (RGB, HSV and grayscale images).
        - A combination of "1", "2", "3" (or equivalently "R", "G", "B" for RGB images):
          Apply the operation to the first/second/third channel (RGB, HSV and grayscale images).
        - "V": Apply the operation to the HSV value (RGB, HSV and and grayscale images).
        - "S": Apply the operation to the HSV saturation (RGB and HSV images).
        - "L": Apply the operation to the luma (RGB and grayscale images).
        - "Ls": Apply the operation to the luma, with highlights protection by desaturation.
          (after the operation, the out-of-range pixels are desaturated at constant luma).
        - "Lb": Apply the operation to the luma, with highlights protection by blending.
          (after the operation, the out-of-range pixels are blended with channels = "RGB").

    Returns:
      Image: The processed image.

    See also:
      skimage.restoration.denoise_bilateral
    """
    if mode == "mirror": # Translate modes.
      mode = "symmetric"
    elif mode == "nearest":
      mode = "edge"
    elif mode == "zero":
      mode = "constant"
    return self.apply_channels(lambda channel: skim.restoration.denoise_bilateral(channel, channel_axis = 0, sigma_spatial = sigma_space,
                               sigma_color = sigma_color, mode = mode, cval = 0.), channels)

  def total_variation(self, weight = .1, algorithm = "Chambolle", channels = ""):
    r"""Total variation denoising of selected channels of the image.

    Given a noisy channel Cin, the total variation filter finds an image Cout with less total variation than Cin
    under the constraint that Cout remains similar to Cin. This can be expressed as the Rudin–Osher–Fatemi (ROF)
    minimization problem:

    .. math::
      \text{Minimize} \sum_{\mathbf{r}} |\nabla C_{out}(\mathbf{r})| + (\lambda/2)[C_{out}(\mathbf{r})-C_{in}(\mathbf{r})]^2

    where the weight 1/lambda controls denoising (the larger the weight, the stronger the denoising at the expense
    of image fidelity). The minimization can either be performed with the Chambolle or split Bregman algorithm.
    Total variation denoising tends to produce cartoon-like (piecewise-constant) images.

    Args:
      weight (float, optional): The weight 1/lambda (default 0.1).
      algorithm (str, optional): Either "Chambolle" (default) for the Chambolle algorithm
                                 or "Bregman" for the split Bregman algorithm.
      channels (str, optional): The selected channels:

        - An empty string (default): Apply the operation to all channels (RGB, HSV and grayscale images).
        - A combination of "1", "2", "3" (or equivalently "R", "G", "B" for RGB images):
          Apply the operation to the first/second/third channel (RGB, HSV and grayscale images).
        - "V": Apply the operation to the HSV value (RGB, HSV and and grayscale images).
        - "S": Apply the operation to the HSV saturation (RGB and HSV images).
        - "L": Apply the operation to the luma (RGB and grayscale images).
        - "Ls": Apply the operation to the luma, with highlights protection by desaturation.
          (after the operation, the out-of-range pixels are desaturated at constant luma).
        - "Lb": Apply the operation to the luma, with highlights protection by blending.
          (after the operation, the out-of-range pixels are blended with channels = "RGB").

    Returns:
      Image: The processed image.

    See also:
      skimage.restoration.denoise_tv_chambolle,
      skimage.restoration.denoise_tv_bregman
    """
    if algorithm == "Chambolle":
      return self.apply_channels(lambda channel: skim.restoration.denoise_tv_chambolle(channel, channel_axis = 0, weight = weight), channels)
    elif algorithm == "Bregman":
      return self.apply_channels(lambda channel: skim.restoration.denoise_tv_bregman(channel, channel_axis = 0, weight = 1./(2.*weight)), channels)
    else:
      raise ValueError(f"Error, unknown algorithm {algorithm} (must be 'Chambolle' or 'Bregman').")

  def non_local_means(self, size = 7, dist = 11, h = .01, sigma = 0., fast = True, channels = ""):
    r"""Non-local means filter for denoising selected channels of the image.

    Given a channel Cin, returns

    .. math::
      C_{out}(\mathbf{r}) \sim \sum_{\mathbf{r}'} f(\mathbf{r},\mathbf{r}') C_{in}(\mathbf{r}')

    where:

    .. math::
      f(\mathbf{r},\mathbf{r}') = \exp[-(M(\mathbf{r})-M(\mathbf{r}'))^2/h^2]\text{ for }|\mathbf{r}-\mathbf{r}'| < d

    and M(r) is an average of Cin in a patch around r.
    Therefore, the non-local means filter averages the neighboring pixels whose patches (texture) are
    sufficiently similar. The non-local means filter can restore textures that would be blurred by
    other denoising algorithms such as the bilateral and total variation filters.

    Args:
      size (int, optional): The size of the (square) patch used to compute M(r) (default 7).
      dist (int, optional): The maximal distance d between the patches (default 11).
      h (float, optional): The cut-off in gray levels (default 0.01; the filter is applied to all channels independently).
      sigma (float, optional): The standard deviation of the noise (if known), subtracted out when computing f(r, r').
        This can lead to a modest improvement in image quality (keep default 0 if unknown).
      fast (bool, optional): If true (default), the pixels within the patch are averaged uniformly.
        If false, they are weighted by a gaussian (better yet slower).
      channels (str, optional): The selected channels:

        - An empty string (default): Apply the operation to all channels (RGB, HSV and grayscale images).
        - A combination of "1", "2", "3" (or equivalently "R", "G", "B" for RGB images):
          Apply the operation to the first/second/third channel (RGB, HSV and grayscale images).
        - "V": Apply the operation to the HSV value (RGB, HSV and and grayscale images).
        - "S": Apply the operation to the HSV saturation (RGB and HSV images).
        - "L": Apply the operation to the luma (RGB and grayscale images).
        - "Ls": Apply the operation to the luma, with highlights protection by desaturation.
          (after the operation, the out-of-range pixels are desaturated at constant luma).
        - "Lb": Apply the operation to the luma, with highlights protection by blending.
          (after the operation, the out-of-range pixels are blended with channels = "RGB").

    Returns:
      Image: The processed image.

    See also:
      skimage.restoration.denoise_nl_means
    """
    return self.apply_channels(lambda channel: skim.restoration.denoise_nl_means(channel, channel_axis = 0, patch_size = size,
                               patch_distance = dist, h = h, sigma = sigma, fast_mode = fast), channels)

  #############
  # Contrast. #
  #############

  def CLAHE(self, size = None, clip = .01, nbins = 256, channels = "L"):
    """Contrast Limited Adaptive Histogram Equalization (CLAHE) of selected channels of the image.

    See https://en.wikipedia.org/wiki/Adaptive_histogram_equalization.

    Args:
      size (optional): The size of the tiles (in pixels) used to sample local histograms,
        given as a single integer or as pair of integers (width, height of the tiles).
        If None (default), the tile size defaults to 1/8 of the image width and height.
      clip (float, optional): The clip limit used to control contrast enhancement (default 0.01).
      nbins (int, optional): The number of bins in the local histograms (default 256).
      channels (str, optional): The selected channels:

        - An empty string (default): Apply the operation to all channels (RGB, HSV and grayscale images).
        - A combination of "1", "2", "3" (or equivalently "R", "G", "B" for RGB images):
          Apply the operation to the first/second/third channel (RGB, HSV and grayscale images).
        - "V": Apply the operation to the HSV value (RGB, HSV and and grayscale images).
        - "S": Apply the operation to the HSV saturation (RGB and HSV images).
        - "L": Apply the operation to the luma (RGB and grayscale images).
        - "Ls": Apply the operation to the luma, with highlights protection by desaturation.
          (after the operation, the out-of-range pixels are desaturated at constant luma).
        - "Lb": Apply the operation to the luma, with highlights protection by blending.
          (after the operation, the out-of-range pixels are blended with channels = "RGB").

        However, CLAHE is only used, in principle, for the "V" (default) and "L(s, b)" channels.

    Returns:
      Image: The processed image.

    See also:
      skimage.exposure.equalize_adapthist
    """
    clipped = self.clip_channels(channels) # Clip relevant channels before CLAHE to avoid artifacts.
    return clipped.apply_channels(lambda channel: skim.exposure.equalize_adapthist(channel, kernel_size = size, clip_limit = clip, nbins = nbins),
                                  channels, multi = False)



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.0.0 / 2024.12.15
# Sphinx OK.

"""Dash backend for Jupyter-lab interface."""

# TODO:
#  - Update tabs only if necessary.

import os
import threading
import numpy as np
import dash
from dash import Dash, dcc, html, ctx
import dash_bootstrap_templates as dbt
import dash_bootstrap_components as dbc
import dash_extensions as dxt

from . import params
from .utils import prepare_images, prepare_images_as_b64strings, shadowed, highlighted, differences
from .backend_plotly import _figure_prepared_image_, _figure_histograms_

from ..equimage import Image, load_image, get_RGB_luma

class Dashboard():
  """Dashboad class."""

  ################
  # Constructor. #
  ################

  def __init__(self, interval = 500, debug = False):
    """Initialize dashboard.

    This dashboard uses Dash to display images, histograms, statistics, etc... in a
    separate browser tab or window.
    It fetches updates from the Dash server at given intervals.

    Args:
      interval (int, optional): The time interval (ms) between dashboard updates (default 500).
      debug (bool, optional): If True, run Dash in debug mode (default False).
    """
    from .. import __packagepath__
    from dash.dependencies import Input, Output, State, ALL, MATCH
    # Initialize object data.
    self.content = []
    self.nupdates = 0
    self.layout = None
    self.images = None
    self.refresh = False
    self.synczoom = False
    self.interval = interval
    self.updatelock = threading.Lock()
    # Set-up Dash app.
    dbt.load_figure_template("slate")
    self.app = Dash(name = __name__, title = "eQuimageLab dashboard", update_title = None,
                    external_stylesheets = [dbc.themes.SLATE], suppress_callback_exceptions = True)
    self.app.layout = self.__layout_dashboard
    # Register callbacks:
    #   - Dashboard update:
    self.app.callback(Output("dashboard", "children"), Output("updateid", "data"),
                      Input("updateinterval", "n_intervals"),
                      running = [Output("updateinterval", "disabled"), True, False], prevent_initial_call = True)(self.__update_dashboard)
    #   - Image click:
    self.app.callback(Output({"type": "datadiv", "index": MATCH}, "children", allow_duplicate = True),
                      Input({"type": "image", "index": MATCH}, "clickData"),
                      State("updateid", "data"),
                      prevent_initial_call = True)(self.__click_image)
    #   - Image filters:
    self.app.callback(Output({"type": "filters", "index": MATCH}, "value"), Output({"type": "selectedfilters", "index": MATCH}, "data"),
                      Output({"type": "image", "index": MATCH}, "figure", allow_duplicate = True),
                      Input({"type": "filters", "index": MATCH}, "value"),
                      State({"type": "selectedfilters", "index": MATCH}, "data"), State("updateid", "data"),
                      prevent_initial_call = True)(self.__filter_image)
    #   - Image histograms:
    self.app.callback(Output({"type": "offcanvas", "index": MATCH}, "is_open"), Output({"type": "offcanvas", "index": MATCH}, "children"),
                      Input({"type": "histogramsbutton", "index": MATCH}, "n_clicks"),
                      State({"type": "offcanvas", "index": MATCH}, "is_open"), State({"type": "image", "index": MATCH}, "figure"),
                      State("updateid", "data"),
                      prevent_initial_call = True)(self.__histograms)
    #   - Images zoom synchronization:
    self.app.callback(Output({"type": "image", "index": ALL}, "relayoutData"), Output({"type": "image", "index": ALL}, "figure"),
                      Input({"type": "image", "index": ALL}, "relayoutData"),
                      prevent_initial_call = True)(self.__sync_image_zoom)
    #   - Histograms zoom tracking:
    self.app.callback(Input({"type": "histograms", "index": ALL}, "relayoutData"), prevent_initial_call = True)(self.__track_histograms_zoom)
    #   - Tab switch.
    self.app.callback(Input("image-tabs", "active_tab"), prevent_initial_call = True)(self.__switch_tab)
    # Launch Dash server.
    self.app.run_server(debug = debug, use_reloader = False, jupyter_mode = "external")
    # Display splash image.
    try:
      splash, meta = load_image(os.path.join(__packagepath__, "images", "splash.png"), verbose = False)
    except:
      pass
    else:
      self.show({"Welcome": splash}, filters = False, click = False, synczoom = False)

  ##############
  # Callbacks. #
  ##############

  def __layout_dashboard(self):
    """Lay out dashboard."""
    with self.updatelock: # Lock while updating layout.
      dashboard = html.Div(self.content, id = "dashboard", className = "dashboard-inner")
      updateid = dcc.Store(data = self.nupdates, id = "updateid")
      interval = dcc.Interval(interval = self.interval, n_intervals = 0, id = "updateinterval")
      return html.Div([dashboard, updateid, interval], className = "dashboard-outer")

  def __update_dashboard(self, n_intervals):
    """Callback for dashboard updates.

    Args:
      n_intervals: The number of update intervals elapsed since the start of the application.

    Returns:
      The updated dashboard and the unique ID of the update.
    """
    if not self.refresh or n_intervals <= 0: return dash.no_update, dash.no_update
    with self.updatelock: # Lock on callback.
      self.refresh = False
      return self.content, self.nupdates

  def __click_image(self, click, updateid):
    """Callback for image click.

    Prints image coordinates and data at click point.

    Args:
      click (dict): The click event dictionary.
      updateid (integer): The unique ID of the displayed dashboard update.

    Returns:
      The content of the "datadiv" div element with the image coordinates and data at click point.
    """
    trigger = ctx.triggered_id # Get the component that triggered the callback.
    if not trigger: return []
    with self.updatelock: # Lock on callback.
      if self.images is None or updateid != self.nupdates: return [] # The dashboard is out of sync.
      n = trigger["index"] # Image index.
      x = click["points"][0]["x"]
      y = click["points"][0]["y"]
      data = self.images[n][y, x]
      if data.size > 1:
        rgbluma = get_RGB_luma()
        luma = rgbluma[0]*data[0]+rgbluma[1]*data[1]+rgbluma[2]*data[2]
        return [f"Data at (x = {x}, y = {y}): ", html.Span(f"R = {data[0]:.5f}", className = "red"), ", ",
                                                 html.Span(f"G = {data[1]:.5f}", className = "green"), ", ",
                                                 html.Span(f"B = {data[2]:.5f}", className = "blue"), ", ",
                                                 html.Span(f"L = {luma:.5f}", className = "luma"), "."]
      else:
        return [f"Data at (x = {x}, y = {y}): ", html.Span(f"L = {data:.5f}", className = "luma"), "."]

  def __filter_image(self, current, previous, updateid):
    """Callback for image filters.

    Apply selected filters (R, G, B, L channel filters, shadowed/highlighted pixels, images differences) to
    the current image.

    Args:
      current (list): The currently selected filters.
      previous (list): The previously selected filters.
      updateid (integer): The unique ID of the displayed dashboard update.

    Returns:
      The curated filters (twice, as currently selected and new previous), and the filtered figure.
    """

    def filter_channels(image, channels):
      """Apply channel filters to the input image."""
      if image.ndim > 2:
        if "L" in channels: # Return luma.
          rgbluma = get_RGB_luma()
          return rgbluma[0]*image[:, : , 0]+rgbluma[1]*image[:, : , 1]+rgbluma[2]*image[:, : , 2]
        else: # Filter out R, G, B channels.
          output = image.copy()
          if "R" not in channels: output[:, :, 0] = 0.
          if "G" not in channels: output[:, :, 1] = 0.
          if "B" not in channels: output[:, :, 2] = 0.
        return output
      else: # Return luma.
        return image

    trigger = ctx.triggered_id # Get the component that triggered the callback.
    if not trigger: return previous, previous, dash.no_update
    with self.updatelock: # Lock on callback.
      if self.images is None or updateid != self.nupdates: return [], [], dash.no_update # The dashboard is out of sync.
      # Update filters list.
      current = set(current)
      previous = set(previous)
      toggled = current^previous
      for t in toggled: # There shall be only one checkbox toggled, actually.
        if t == "L":
          if "L" in current:
            current.difference_update({"R", "G", "B"})
          else:
            current.update({"R", "G", "B"})
        elif t in ["R", "G", "B"]:
          if not current & {"R", "G", "B"}:
            current.update("L")
          else:
            current.difference_update({"L"})
        elif t == "S":
          current.difference_update({"H", "D"})
        elif t == "H":
          current.difference_update({"S", "D"})
        elif t == "D":
          current.difference_update({"S", "H"})
        else:
          raise ValueError(f"Error, unknown filter {t}.")
      # Apply selected filters to the image.
      n = trigger["index"] # Image index.
      image = filter_channels(self.images[n], current)
      if current & {"S", "H", "D"}:
        reference = filter_channels(self.images[self.reference], current) if self.reference is not None else None
        if "S" in current:
          image = shadowed(image, reference)
        elif "H" in current:
          image = highlighted(image, reference)
        else:
          image = differences(image, reference)
      # Return filtered image as a patch.
      figure_patch = dash.Patch()
      figure_patch["data"][0]["source"] = prepare_images_as_b64strings(image, sampling = 1)
      current = list(current)
      return current, current, figure_patch

  def __histograms(self, n_clicks, is_open, figure, updateid):
    """Callback for local histograms.

    Shows the histograms of the displayed area of the current image.

    Args:
      n_clicks (list): The number of clicks on the histograms button.
      is_open (bool): The status of the off-canvas showing the local histograms.
      figure (dict): The figure associated with the histograms button.
      updateid (integer): The unique ID of the displayed dashboard update.

    Returns:
      The status and content of the off-canvas showing the local histograms.
    """
    if is_open: return False, [] # Close the off-canvas.
    if n_clicks <= 0: return False, []
    trigger = ctx.triggered_id # Get the component that triggered the callback.
    if not trigger: return False, []
    with self.updatelock: # Lock on callback.
      if self.images is None or updateid != self.nupdates: return False, [] # The dashboard is out of sync.
      # Get x and y axes range.
      xmin, xmax = figure["layout"]["xaxis"]["range"]
      xmin = np.ceil(xmin) ; xmax = np.floor(xmax)
      ymax, ymin = figure["layout"]["yaxis"]["range"]
      ymin = np.ceil(ymin) ; ymax = np.floor(ymax)
      if (xmax-xmin)*(ymax-ymin) < 256: return False, [] # Area too small for histograms.
      # Compute local histograms using eQuimage.
      n = trigger["index"] # Image index.
      image = Image(self.images[n], channels = -1).crop(xmin, xmax, ymin, ymax)
      figure = _figure_histograms_(image, channels = "RGBL", log = True, width = params.maxwidth, template = "slate")
      content = [dcc.Graph(figure = figure)]
      return True, content

  def __sync_image_zoom(self, relayouts):
    """Callback for images zoom synchronization.

    Args:
      relayouts (list): Input relayouts of all images.

    Returns:
      Output relayouts and figure patches for all images.
    """
    nimages = len(relayouts)
    if not self.synczoom: return [dash.no_update]*nimages, [dash.no_update]*nimages
    trigger = ctx.triggered_id # Get the component that triggered the callback.
    if not trigger: return [dash.no_update]*nimages, [dash.no_update]*nimages
    with self.updatelock: # Lock on callback.
      n = trigger["index"] # Image index.
      relayout = relayouts[n]
      xauto = relayout.get("xaxis.autorange", False)
      if not xauto:
        xmin = relayout.get("xaxis.range[0]", None)
        xmax = relayout.get("xaxis.range[1]", None)
        if xmin is None or xmax is None:
          return [dash.no_update]*nimages, [dash.no_update]*nimages # Unexpected relayout structure; Discard event.
      yauto = relayout.get("yaxis.autorange", False)
      if not yauto:
        ymin = relayout.get("yaxis.range[0]", None)
        ymax = relayout.get("yaxis.range[1]", None)
        if ymin is None or ymax is None:
          return [dash.no_update]*nimages, [dash.no_update]*nimages # Unexpected relayout structure; Discard event.
      figure_patch = dash.Patch()
      figure_patch["layout"]["xaxis"]["autorange"] = xauto
      figure_patch["layout"]["yaxis"]["autorange"] = yauto
      if xauto:
        self.xrange = None
      else:
        self.xrange = [xmin, xmax]
        figure_patch["layout"]["xaxis"]["range"] = self.xrange
      if yauto:
        self.yrange = None
      else:
        self.yrange = [ymin, ymax]
        figure_patch["layout"]["yaxis"]["range"] = self.yrange
      return [relayout]*nimages, [figure_patch]*nimages

  def __track_histograms_zoom(self, relayouts):
    """Callback for histograms zoom tracking.

    Tracks x axis changes on histograms.

    Args:
      relayouts (dict): Input relayouts of all histograms.
    """
    if not self.synczoom: return # This comes along with image zoom synchronization.
    trigger = ctx.triggered_id # Get the component that triggered the callback.
    if not trigger: return
    with self.updatelock: # Lock on callback.
      n = trigger["index"] # Image index.
      relayout = relayouts[n]
      xauto = relayout.get("xaxis.autorange", False)
      if xauto:
        self.hrange[n] = None
      else:
        xmin = relayout.get("xaxis.range[0]", None)
        xmax = relayout.get("xaxis.range[1]", None)
        if xmin is None or xmax is None: return # Unexpected relayout structure; Discard event.
        self.hrange[n] = [xmin, xmax]

  def __switch_tab(self, tab):
    """Callback for tab switches.

    This callback just stores the current tab name.

    Args:
      tab (string): The name of the current tab.
    """
    with self.updatelock: # Lock on callback.
      self.activetab = tab

  ############
  # Layouts. #
  ############

  ### Tabs layout.

  def show(self, images, histograms = False, statistics = False, sampling = -1, filters = True, click = True, synczoom = True, trans = None):
    """Show image(s) on the dashboard.

    Args:
      images: The image(s) as a single/tuple/list/dictionary of Image object(s) or numpy.ndarray.
        Each image is displayed in a separate tab. The tabs are labelled according to the keys for
        a dictionary. Otherwise, the tabs are labelled "Image" & "Reference" if there are one or two images,
        and "Image #1", "Image #2"... if there are more.
        The images must all be Image objects if histograms or statistics is True.
      histograms (optional): If True or a string, show the histograms of the image(s). The string lists the
        channels of the histograms (e.g. "RGBL" for red, green, blue, luma). Default is False.
      statistics (optional): If True or a string, show the statistics of the image(s). The string lists the
        channels of the statistics (e.g. "RGBL" for red, green, blue, luma). Default is False.
      sampling (int, optional): Downsampling rate (defaults to `jupyter.params.sampling` if negative).
        Only images[::sampling, ::sampling] are shown, to speed up display.
      filters (bool, optional): If True (default), add image filters menu (R, G, B, L channel filters,
        shadowed/highlighted pixels, images differences, local histograms).
      click (bool, optional): If True, show image data on click (default True).
      synczoom (bool, optional): If True (default), synchronize zooms over images.
        Zooms will be synchronized only if all images have the same size.
      trans (optional): A container with an histogram transformation (see Image.apply_channels), plotted on
        top of the histograms of the "Reference" tab (default None).
    """
    self.refresh = False # Stop refreshing dashboard.
    # Build the dictionary of images.
    if isinstance(images, (tuple, list)):
      nimages = len(images)
      if nimages == 1:
        keys = ["Image"]
      elif nimages == 2:
        keys = ["Image", "Reference"]
      else:
        keys = [f"Image #{n+1}" for n in range(nimages)]
    elif isinstance(images, dict):
      nimages = len(images)
      keys = list(images.keys())
      images = list(images.values())
    else:
      nimages = 1
      keys = ["Image"]
      images = [images]
    # Look for a reference image.
    try:
      reference = keys.index("Reference")
    except:
      reference = None
    # Prepare images.
    pimages = prepare_images(images, sampling = sampling)
    # Check if zooms can be synchronized.
    if synczoom:
      imagesize = pimages[0].shape[0:2]
      for image in pimages[1:]:
        synczoom = (image.shape[0:2] == imagesize)
        if not synczoom: break
    if not synczoom: imagesize = None
    # Try to preserve existing axes if already in tabs layout and synczoom is consistently True.
    preserveaxes = synczoom and self.layout == "tabs" and self.synczoom and self.imagesize == imagesize
    if preserveaxes:
      xrange = self.xrange # Image x & y ranges.
      yrange = self.yrange
    else:
      xrange = None
      yrange = None
    hrange = [None]*nimages # Histograms x ranges.
    # Set-up tabs.
    tabs = []
    for n in range(nimages):
      tab = []
      figure = _figure_prepared_image_(pimages[n], width = params.maxwidth, hover = False, template = "slate")
      if xrange is not None: figure.update_layout(xaxis_range = xrange)
      if yrange is not None: figure.update_layout(yaxis_range = yrange)
      if click: figure.update_layout(clickmode = "event+select")
      tab.append(dcc.Graph(figure = figure, id = {"type": "image", "index": n}))
      if filters:
        options = []
        values = []
        if pimages[n].ndim > 2: # Color image.
          options.extend([{"label": html.Span("R", className = "red lm1"), "value": "R"},
                          {"label": html.Span("G", className = "green lm1"), "value": "G"},
                          {"label": html.Span("B", className = "blue lm1"), "value": "B"},
                          {"label": html.Span("L", className = "luma lm1 rm4"), "value": "L"}])
          values.extend(["R", "G", "B"])
        options.extend([{"label": html.Span("Shadowed", className = "lm1"), "value": "S"},
                        {"label": html.Span("Highlighted", className = "lm1"), "value": "H"}])
        if reference is not None and pimages[n].shape == pimages[reference].shape:
          options.extend([{"label": html.Span("Differences", className = "lm1"), "value": "D"}])
        checklist = dcc.Checklist(options = options, value = values, id = {"type": "filters", "index": n},
                                  inline = True, labelClassName = "rm4")
        selected = dcc.Store(data = values, id = {"type": "selectedfilters", "index": n})
        button = dbc.Button("Local histograms", color = "primary", size = "sm", n_clicks = 0, id = {"type": "histogramsbutton", "index": n})
        offcanvas = dbc.Offcanvas([], title = "Histograms of the displayed area of the image:", placement = "top", close_button = True, keyboard = True,
                                  id = {"type": "offcanvas", "index": n}, style = {"height": "auto", "bottom": "initial"}, is_open = False)
        tab.append(html.Div([html.Div(["Filters:"], className = "rm4"), html.Div([checklist]),  html.Div([button], className = "flushright"),
                             selected, offcanvas], className = "flex center tm1 bm1",
                             style = {"width": f"{params.maxwidth}px", "margin-left": f"{params.lmargin}px", "margin-right": f"{params.rmargin}px"}))
      if click:
        tab.append(html.Div([], id = {"type": "datadiv", "index": n}, className = "tm1 bm1",
                   style = {"width": f"{params.maxwidth}px", "margin-left": f"{params.lmargin}px", "margin-right": f"{params.rmargin}px"}))
      if histograms is not False:
        if histograms is True: histograms = ""
        figure = _figure_histograms_(images[n], channels = histograms, log = True, width = params.maxwidth,
                                     trans = trans if keys[n] == "Reference" else None, template = "slate")
        if figure is not None:
          if preserveaxes and keys[n] in self.keys:
            hrange[n] = self.hrange[self.keys.index(keys[n])]
            if hrange[n] is not None: figure.update_layout(xaxis_range = hrange[n])
          tab.append(dcc.Graph(figure = figure, id = {"type": "histograms", "index": n}))
      if statistics is not False:
        if statistics is True: statistics = ""
        table = _table_statistics_(images[n], channels = statistics)
        if table is not None: tab.append(table)
      tabs.append(dbc.Tab(tab, label = keys[n], tab_id = keys[n], className = "tab"))
    # Set active tab. Keep current tab open if possible.
    activetab = self.activetab if self.layout == "tabs" and self.activetab in keys else keys[0]
    # Update dashboard.
    with self.updatelock: # Lock on update.
      # BEWARE TO SIDE EFFECTS: SELF.IMAGES MAY REFERENCE THE ORIGINAL IMAGES.
      self.nupdates += 1
      self.layout = "tabs"
      self.keys = keys
      self.xrange = xrange
      self.yrange = yrange
      self.hrange = hrange
      self.synczoom = synczoom
      self.imagesize = imagesize
      self.reference = reference
      self.activetab = activetab
      self.images = pimages if click or filters else None # No need to register images for the callbacks if click and filters are False.
      self.content = [dbc.Tabs(tabs, active_tab = activetab, id = "image-tabs")]
      self.refresh = True

  def show_t(self, image, channels = "RGBL", sampling = -1, filters = True, click = True, synczoom = True):
    """Show the input and output images of an histogram transformation on the dashboard.

    Displays the input image, histograms, statistics, and transformation curve in tab "Reference",
    and the output image, histograms, and statistics in tab "Image".

    Args:
      image (Image): The output image (must embed a transformation image.trans - see Image.apply_channels).
      channels (str, optional): The channels of the histograms and statistics (default "RGBL" for red,
        green, blue, luma). The channels of the transformation are added if needed.
      sampling (int, optional): Downsampling rate (defaults to `jupyter.params.sampling` if negative).
        Only image[::sampling, ::sampling] is shown, to speed up display.
      filters (bool, optional): If True (default), add image filters menu (R, G, B, L channel filters,
        shadowed/highlighted pixels, images differences, local histograms).
      click (bool, optional): If True, show image data on click (default True).
      synczoom (bool, optional): If True (default), synchronize zooms over images.
    """
    if not issubclass(type(image), Image): print("The transformations can only be displayed for Image objects.")
    trans = getattr(image, "trans", None)
    if trans is None:
      print("There is no transformation embedded in the input image.")
      return
    reference = trans.input
    if trans.type == "hist":
      for c in trans.channels:
        if c in "RGBVSL" and not c in channels:
          channels += c
    self.show({"Image": image, "Reference": reference}, histograms = channels, statistics = channels,
              sampling = sampling, filters = filters, click = click, synczoom = synczoom, trans = trans)

  ### Carousel layout.

  def carousel(self, images, sampling = -1, interval = 2000):
    """Show a carousel of images on the dashboard.

    Args:
      images: The images as a tuple/list/dictionary of Image object(s) or numpy.ndarray.
        The images are labelled according to the keys for a dictionary. Otherwise, the images are labelled
        "Image" and "Reference" if there are two images, and "Image #1", "Image #2"... if there are more.
      sampling (int, optional): Downsampling rate (defaults to `jupyter.params.sampling` if negative).
        Only images[::sampling, ::sampling] are shown, to speed up display.
      interval (int, optional): The interval (ms) between image switches in the carousel (default 2000).
    """
    self.refresh = False # Stop refreshing dashboard.
    # Build the dictionary of images.
    if isinstance(images, (tuple, list)):
      nimages = len(images)
      if nimages == 1:
        keys = ["Image"]
      elif nimages == 2:
        keys = ["Image", "Reference"]
      else:
        keys = [f"Image #{n+1}" for n in range(nimages)]
    elif isinstance(images, dict):
      nimages = len(images)
      keys = list(images.keys())
      images = list(images.values())
    else:
      nimages = 1
      keys = ["Image"]
      images = [images]
    # Set-up carousel.
    items = [dict(key = f"{n}", src = prepare_images_as_b64strings(images[n], sampling = sampling), header = keys[n]) for n in range(nimages)]
    widget = dbc.Carousel(items = items, controls = True, indicators = True, ride = "carousel", interval = interval, className = "carousel-fade",
             style = {"width": f"{params.maxwidth}px", "margin": f"{params.tmargin}px {params.rmargin}px {params.bmargin}px {params.lmargin}px"})
    tab = dbc.Tab([widget], label = "Carousel", className = "tab")
    # Update dashboard.
    with self.updatelock: # Lock on update.
      self.nupdates += 1
      self.layout = "carousel"
      self.images = None # No need to register images for the callbacks.
      self.content = [dbc.Tabs([tab], active_tab = "tab-0")]
      self.refresh = True

  ### Before/After layout.

  def slide(self, image1, image2, label1 = "Image", label2 = "Reference", sampling = -1):
    """Compare two images with a "before/after" slider on the dashboard.

    Args:
      image1: The first image as an Image object or numpy.ndarray.
      image2: The second image as an Image object or numpy.ndarray.
      label1 (str, optional): The label of the first image (default "Image").
      label2 (str, optional): The label of the second image (default "Reference").
      sampling (int, optional): Downsampling rate (defaults to `jupyter.params.sampling` if negative).
        Only image1[::sampling, ::sampling] and image2[::sampling, ::sampling] are shown, to speed up display.
    """
    self.refresh = False # Stop refreshing dashboard.
    # Set-up before/after widget.
    image1, image2 = prepare_images_as_b64strings((image1, image2), sampling = sampling)
    baslider = dxt.BeforeAfter(before = dict(src = image1), after = dict(src = image2), width = f"{params.maxwidth}")
    left   = html.Div([label2], className = "ba-left", style = {"width": f"{params.lmargin}px"})
    middle = html.Div([baslider], className = "ba-middle", style = {"width": f"{params.maxwidth}px"})
    right  = html.Div([label1], className = "ba-right", style = {"width": f"{params.rmargin}px"})
    widget = html.Div([left, middle, right], className = "inline",
                      style = {"margin": f"{params.tmargin}px 0px {params.bmargin}px 0px"})
    tab = dbc.Tab([widget], label = "Compare images", className = "tab")
    # Update dashboard.
    with self.updatelock: # Lock on update.
      self.nupdates += 1
      self.layout = "slider"
      self.images = None # No need to register images for the callbacks.
      self.content = [dbc.Tabs([tab], active_tab = "tab-0")]
      self.refresh = True

#####################
# Helper functions. #
#####################

def _table_statistics_(image, channels = ""):
  """Prepare a table with the statistics of an image.

  Args:
    image (Image): The image.
    channels (str, optional): The channels of the histograms (default "" = "RGBL" for red, green, blue, luma).

  Returns:
    dbc.Table: A dash bootstrap components table with the statistics of the image.
  """
  # Prepare statistics.
  if not issubclass(type(image), Image):
    print("The statistics can only be displayed for Image objects.")
    return None
  if channels == "":
    stats = getattr(image, "stats", None)
    if stats is None: stats = image.statistics()
  else:
    stats = image.statistics(channels = channels)
  # Create table.
  header = html.Thead(html.Tr([html.Th("Channel", className = "left"), html.Th("Minimum"), html.Th("25%"), html.Th("50%"),
                               html.Th("75%"), html.Th("Maximum"), html.Th("Shadowed"), html.Th("Highlighted")]))
  rows = []
  exclude01 = False
  for channel in stats.values():
    exclude01 = exclude01 or channel.exclude01
    deco = "\u207d\u2071\u207e" if channel.exclude01 else ""
    if channel.percentiles is not None:
      percentiles = [f"{channel.percentiles[0]:.5f}{deco}", f"{channel.percentiles[1]:.5f}{deco}", f"{channel.percentiles[2]:.5f}{deco}"]
    else:
      percentiles = 3*[f"None{deco}"]
    rows.append(html.Tr([html.Td(channel.name, className = "bold left", style = {"color": f"{channel.color}"}), html.Td(f"{channel.minimum:.5f}"),
                         html.Td(percentiles[0]), html.Td(percentiles[1]), html.Td(percentiles[2]), html.Td(f"{channel.maximum:.5f}"),
                         html.Td(f"{channel.zerocount} ({100.*channel.zerocount/channel.npixels:.2f}%)"),
                         html.Td(f"{channel.outcount} ({100.*channel.outcount/channel.npixels:.2f}%)")]))
  body = html.Tbody(rows)
  table = [dbc.Table([header]+[body], size = "sm", bordered = True, striped = True, className = "right", style = {"width": f"{params.maxwidth}px"})]
  if exclude01: table.append("\u207d\u2071\u207e Does not include pixels <= 0 or >= 1.")
  return html.Div(table, className = "tm8 bm8", style = {"margin-left": f"{params.lmargin}px", "margin-right": f"{params.rmargin}px"})



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.0.0 / 2024.12.15
# Sphinx OK.

"""Color spaces and models management."""

import numpy as np
import skimage.color as skcolor

from . import params
from . import helpers

#############################
# sRGB <-> lRGB conversion. #
#############################

def sRGB_to_lRGB(image):
  """Convert the input sRGB image into a linear RGB image.

  See also:
    The reciprocal lRGB_to_sRGB function.

  Args:
    image (numpy.ndarray): The input sRGB image.

  Returns:
    numpy.ndarray: The converted lRGB image.
  """
  srgb = np.clip(image, 0., None)
  return np.where(srgb > .04045, ((srgb+0.055)/1.055)**2.4, srgb/12.92)

def lRGB_to_sRGB(image):
  """Convert the input linear RGB image into a sRGB image.

  See also:
    The reciprocal sRGB_to_lRGB function.

  Args:
    image (numpy.ndarray): The input lRGB image.

  Returns:
    numpy.ndarray: The converted sRGB image.
  """
  lrgb = np.clip(image, 0., None)
  return np.where(lrgb > .0031308, 1.055*lrgb**(1./2.4)-0.055, 12.92*lrgb)

###########################
# RGB <-> HSV conversion. #
###########################

def RGB_to_HSV(image):
  """Convert the input RGB image into a HSV image.

  See also:
    The reciprocal HSV_to_RGB function.

  Args:
    image (numpy.ndarray): The input RGB image.

  Returns:
    numpy.ndarray: The converted HSV image.
  """
  return skcolor.rgb2hsv(image, channel_axis = 0)

def HSV_to_RGB(image):
  """Convert the input HSV image into a RGB image.

  See also:
    The reciprocal RGB_to_HSV function.

  Args:
    image (numpy.ndarray): The input HSV image.

  Returns:
    numpy.ndarray: The converted RGB image.
  """
  return skcolor.hsv2rgb(image, channel_axis = 0)

def value(image):
  """Return the HSV value V = max(RGB) of the input RGB image.

  Note: Compatible with single channel grayscale images.

  Args:
    image (numpy.ndarray): The input RGB image.

  Returns:
    numpy.ndarray: The HSV value V.
  """
  return image.max(axis = 0)

def saturation(image):
  """Return the HSV saturation S = 1-min(RGB)/max(RGB) of the input RGB image.

  Note: Compatible with single channel grayscale images.

  Args:
    image (numpy.ndarray): The input RGB image.

  Returns:
    numpy.ndarray: The HSV saturation S.
  """
  return 1.-image.min(axis = 0)/image.max(axis = 0, initial = params.IMGTOL) # Safe evaluation.

#########
# Luma. #
#########

def luma(image):
  """Return the luma L of the input RGB image.

  The luma L is the average of the RGB components weighted by rgbluma = get_RGB_luma():

    L = rgbluma[0]*image[0]+rgbluma[1]*image[1]+rgbluma[2]*image[2].

  Note: Compatible with single channel grayscale images.

  Args:
    image (numpy.ndarray): The input RGB image.

  Returns:
    numpy.ndarray: The luma L.
  """
  rgbluma = params.get_RGB_luma()
  return rgbluma[0]*image[0]+rgbluma[1]*image[1]+rgbluma[2]*image[2] if image.shape[0] > 1 else image[0]

############################
# Luminance and lightness. #
############################

def lRGB_luminance(image):
  """Return the luminance Y of the input linear RGB image.

  The luminance Y of a lRGB image is defined as:

    Y = 0.2126*R+0.7152*G+0.0722*B

  It is equivalently the luma of the lRGB image for RGB weights (0.2126, 0.7152, 0.0722),
  and is the basic ingredient of the perceptual lightness L*.

  Note: Compatible with single channel grayscale images.

  See also:
    lRGB_lightness,
    sRGB_luminance,
    sRGB_lightness,
    luma

  Args:
    image (numpy.ndarray): The input lRGB image.

  Returns:
    numpy.ndarray: The luminance Y.
  """
  return .2126*image[0]+.7152*image[1]+.0722*image[2] if image.shape[0] > 1 else image[0]

def lRGB_lightness(image):
  """Return the CIE lightness L* of the input linear RGB image.

  The CIE lightness L* is defined from the lRGB luminance Y as:

    L* = 116*Y**(1/3)-16 if Y > 0.008856 and L* = 903.3*Y if Y < 0.008856.

  It is a measure of the perceptual lightness of the image.

  Warning: L* is defined within [0, 100] instead of [0, 1].

  Note: Compatible with single channel grayscale images.

  See also:
    lRGB_luminance,
    sRGB_luminance,
    sRGB_lightness,
    luma

  Args:
    image (numpy.ndarray): The input lRGB image.

  Returns:
    numpy.ndarray: The CIE lightness L*.
  """
  Y = lRGB_luminance(image)
  return np.where(Y > .008856, 116.*Y**(1./3.)-16., 903.3*Y)

def sRGB_luminance(image):
  """Return the luminance Y of the input sRGB image.

  The image is converted to the lRGB color space to compute the luminance Y.

  Note: Although they have the same functional forms, the luma and luminance are
  different concepts for sRGB images: the luma is computed in the sRGB color space
  as a *substitute* for the perceptual lightness, whereas the luminance is
  computed after conversion in the lRGB color space and is the basic ingredient of
  the *genuine* perceptual lightness (see lRGB_lightness).

  Note: Compatible with single channel grayscale images.

  See also:
    sRGB_lightness,
    lRGB_luminance,
    lRGB_lightness,
    luma

  Args:
    image (numpy.ndarray): The input sRGB image.

  Returns:
    numpy.ndarray: The luminance Y.
  """
  return lRGB_luminance(sRGB_to_lRGB(image))

def sRGB_lightness(image):
  """Return the CIE lightness L* of the input sRGB image.

  The image is converted to the lRGB color space to compute the CIE lightness L*.
  L* is a measure of the perceptual lightness of the image.

  Warning: L* is defined within [0, 100] instead of [0, 1].

  Note: Compatible with single channel grayscale images.

  See also:
    sRGB_luminance,
    lRGB_luminance,
    lRGB_lightness,
    luma

  Args:
    image (numpy.ndarray): The input sRGB image.

  Returns:
    numpy.ndarray: The CIE lightness L*.
  """
  return lRGB_lightness(sRGB_to_lRGB(image))

#####################################
# For inclusion in the Image class. #
#####################################

class Mixin:
  """To be included in the Image class."""

  #####################################
  # Color space and model management. #
  #####################################

  def color_space_error(self):
    """Raise a color space error."""
    class ColorSpaceError(Exception):
      pass
    raise ColorSpaceError(f"Error, this operation is not implemented for {self.colorspace} images.")

  def color_model_error(self):
    """Raise a color model error."""
    class ColorModelError(Exception):
      pass
    raise ColorModelError(f"Error, this operation is not implemented for {self.colormodel} images.")

  def check_color_space(self, *colorspaces):
    """Raise a color space error if the color space of the image is not in the arguments.

    See also:
      color_space_error
    """
    if self.colorspace not in colorspaces: self.color_space_error()

  def check_color_model(self, *colormodels):
    """Raise a color model error if the color model of the image is not in the arguments.

    See also:
      color_model_error
    """
    if self.colormodel not in colormodels: self.color_model_error()

  def check_is_not_gray(self):
    """Raise a color model error if the image is a grayscale.

    See also:
      color_model_error
    """
    if self.colormodel == "gray": self.color_model_error()

  ############################
  # Color space conversions. #
  ############################

  def lRGB(self):
    """Convert the image to the linear RGB color space.

    Warning:
      The color model must be "RGB" or "gray".

    Returns:
      Image: The converted lRGB image (a copy of the original image if already lRGB).
    """
    self.check_color_model("RGB", "gray")
    if self.colorspace == "lRGB":
      return self.copy()
    elif self.colorspace == "sRGB":
      return self.newImage(sRGB_to_lRGB(self.image), colorspace = "lRGB")
    else:
      self.color_space_error()

  def sRGB(self):
    """Convert the image to the sRGB color space.

    Warning:
      The color model must be "RGB" or "gray".

    Returns:
      Image: The converted sRGB image (a copy of the original image if already sRGB).
    """
    self.check_color_model("RGB", "gray")
    if self.colorspace == "lRGB":
      return self.newImage(lRGB_to_sRGB(self.image), colorspace = "sRGB")
    elif self.colorspace == "sRGB":
      return self.copy()
    else:
      self.color_space_error()

  ############################
  # Color model conversions. #
  ############################

  def RGB(self):
    """Convert the image to the RGB color model.

    Returns:
      Image: The converted RGB image (a copy of the original image if already RGB).
    """
    if self.colormodel == "RGB":
      return self.copy()
    elif self.colormodel == "HSV":
      return self.newImage(HSV_to_RGB(self.image), colormodel = "RGB")
    elif self.colormodel == "gray":
      return self.newImage(np.repeat(self.image[0, :, :], 3, axis = 0), colormodel = "RGB")
    else:
      self.color_model_error()

  def HSV(self):
    """Convert the image to the HSV color model.

    Warning:
      The conversion from a gray scale to a HSV image is ill-defined (no hue).

    Returns:
      Image: The converted HSV image (a copy of the original image if already HSV).
    """
    if self.colormodel == "RGB":
      return self.newImage(RGB_to_HSV(self.image), colormodel = "HSV")
    elif self.colormodel == "HSV":
      return self.copy()
    else:
      self.color_model_error()

  #######################
  # Composite channels. #
  #######################

  def value(self):
    """Return the HSV value V = max(RGB) of the image.

    Returns:
      numpy.ndarray: The HSV value V.
    """
    if self.colormodel == "RGB" or self.colormodel == "gray":
      return value(self.image)
    elif self.colormodel == "HSV":
      return self.image[2]
    else:
      self.color_model_error()

  def saturation(self):
    """Return the HSV saturation S = 1-min(RGB)/max(RGB) of the image.

    Returns:
      numpy.ndarray: The HSV saturation S.
    """
    if self.colormodel == "RGB" or self.colormodel == "gray":
      return saturation(self.image)
    elif self.colormodel == "HSV":
      return self.image[1]
    else:
      self.color_model_error()

  def luma(self):
    """Return the luma L of the image.

    The luma L is the average of the RGB components weighted by rgbluma = get_RGB_luma():

      L = rgbluma[0]*image[0]+rgbluma[1]*image[1]+rgbluma[2]*image[2].

    Warning:
      The luma is available only for RGB and grayscale images.

    Returns:
      numpy.ndarray: The luma L.
    """
    if self.colormodel == "RGB" or self.colormodel == "gray":
      return luma(self.image)
    else:
      self.color_model_error()

  def luminance(self):
    """Return the luminance Y of the image.

    Warning:
      The luminance is available only for RGB and grayscale images.

    Returns:
      numpy.ndarray: The luminance Y.
    """
    self.check_color_model("RGB", "gray")
    if self.colorspace == "lRGB":
      return lRGB_luminance(self.image)
    elif self.colorspace == "sRGB":
      return sRGB_luminance(self.image)
    else:
      self.color_space_error()

  def lightness(self):
    """Return the CIE lightness L* of the image.

    Warning:
      The lightness is available only for RGB and grayscale images.

    Warning:
      L* is defined within [0, 100] instead of [0, 1].

    Returns:
      numpy.ndarray: The lightness L*.
    """
    self.check_color_model("RGB", "gray")
    if self.colorspace == "lRGB":
      return lRGB_lightness(self.image)
    elif self.colorspace == "sRGB":
      return sRGB_lightness(self.image)
    else:
      raise self.color_space_error()

  #################################
  # Channel-selective operations. #
  #################################

  def apply_channels(self, f, channels, multi = True, trans = False):
    """Apply the operation f(channel) to selected channels of the image.

    Note: When applying an operation to the luma, the RGB components of the image are rescaled
    by the ratio f(luma)/luma. This preserves the hue, but may bring some RGB components
    out-of-range even though f(luma) is within [0, 1]. These out-of-range components can be
    regularized with two highlights protection methods:

      - "saturation": The out-of-range pixels are desaturated at constant luma (namely, the
        out-of-range components are decreased while the in-range components are increased so
        that the luma is conserved). This tends to whiten the out-of-range pixels.
        f(luma) must be within [0, 1] to make use of this highlights protection method.
      - "mixing": The out-of-range pixels are blended with f(RGB) (the operation applied to the
        RGB channels). This usually tends to whiten the out-of-range pixels too.
        f(RGB) must be within [0, 1] to make use of this highlights protection method.

    Alternatively, applying the operation to the HSV value V also preserves the hue and can not
    produce out-of-range pixels if f([0, 1]) is within [0, 1]. However, this may strongly affect
    the balance of the image, the HSV value being a very poor approximation to the perceptual
    lightness.

    Args:
      f (function): The function f(numpy.ndarray) → numpy.ndarray applied to the selected channels.
      channels (str): The selected channels:

        - An empty string: Apply the operation to all channels (RGB, HSV and grayscale images).
        - A combination of "1", "2", "3" (or equivalently "R", "G", "B" for RGB images):
          Apply the operation to the first/second/third channel (RGB, HSV and grayscale images).
        - "V": Apply the operation to the HSV value (RGB, HSV and and grayscale images).
        - "S": Apply the operation to the HSV saturation (RGB and HSV images).
        - "L": Apply the operation to the luma (RGB and grayscale images).
        - "Ls": Apply the operation to the luma, with highlights protection by desaturation.
          (after the operation, the out-of-range pixels are desaturated at constant luma).
        - "Lb": Apply the operation to the luma, with highlights protection by blending.
          (after the operation, the out-of-range pixels are blended with f(RGB)).

      multi (bool, optional): if True (default), the operation can be applied to the whole image at once;
        if False, the operation must be applied one channel at a time.
      trans (bool, optional): If True (default False), embeds the transformation y = f(x in [0, 1]) in the
        output image as output.trans, where:

          - output.trans.type = "hist".
          - output.trans.input is a reference to the input image (self).
          - output.trans.channels are the channels selected for the transformation.
          - output.trans.x is a mesh of the [0, 1] interval.
          - output.trans.y = f(output.trans.x)
          - output.trans.ylabel is a label for output.trans.y.
          - output.trans.xticks is a list of remarkable x values for this transformation (if any).

        trans shall be set True only for *local* histogram transformations f.

    Returns:
      Image: The processed image.
    """

    def transformation(f, x, channels):
      """Return the transformation container."""
      xmin = min(0., x.min())
      xmax = max(1., x.max())
      t = helpers.Container()
      t.type = "hist"
      t.input = self
      t.channels = channels
      t.x = np.linspace(xmin, xmax, max(int(round(params.ntranshi*(xmax-xmin))), 2*params.ntranshi))
      t.y = f(t.x)
      t.ylabel = "f"
      return t

    is_RGB  = (self.colormodel == "RGB")
    is_HSV  = (self.colormodel == "HSV")
    is_gray = (self.colormodel == "gray")
    if channels == "":
      if is_gray:
        channels = "L"
      elif is_RGB:
        channels = "RGB"
      else:
        channels = "123"
    if channels == "V":
      if is_gray:
        output = self.newImage(f(self.image))
        if trans: output.trans = transformation(f, self.image, "V")
        return output
      elif is_RGB:
        value = self.value()
        output = self.scale_pixels(value, f(value))
        if trans: output.trans = transformation(f, value, "V")
        return output
      elif is_HSV:
        output = self.copy()
        output.image[2] = f(self.image[2])
        if trans: output.trans = transformation(f, self.image[2], "V")
        return output
      else:
        self.color_model_error()
    elif channels == "S":
      if is_RGB:
        hsv = self.HSV()
        if trans: t = transformation(f, hsv.image[1], "S")
        hsv.image[1] = f(hsv.image[1])
        output = hsv.RGB()
        if trans: output.trans = t
        return output
      elif is_HSV:
        output = self.copy()
        output.image[1] = f(self.image[1])
        if trans: output.trans = transformation(f, self.image[1], "S")
        return output
      else:
        self.color_model_error()
    elif channels == "L" or channels == "Ls" or channels == "Lb":
      if is_gray:
        output = self.newImage(f(self.image))
        if trans: output.trans = transformation(f, self.image, "L")
        return output
      elif is_RGB:
        luma = self.luma()
        output = self.scale_pixels(luma, f(luma))
        if channels == "Ls":
          output = output.protect_highlights_sat()
        elif channels == "Lb":
          output = output.protect_highlights_blend(self.apply_channels(f, "RGB", multi))
        if trans: output.trans = transformation(f, luma, "L")
        return output
      else:
        self.color_model_error()
    else:
      nc = self.get_nc()
      selected = nc*[False]
      for c in channels:
        ok = True
        if c == "1":
          ic = 0
        elif c == "2":
          ic = 1
          ok = not is_gray
        elif c == "3":
          ic = 2
          ok = not is_gray
        elif c == "R":
          ic = 0
          ok = is_RGB
        elif c == "G":
          ic = 1
          ok = is_RGB
        elif c == "B":
          ic = 2
          ok = is_RGB
        else:
          ok = False
        if not ok: raise ValueError(f"Error, unknown or incompatible channel '{c}'.")
        if selected[ic]: print(f"Warning, channel '{c}' selected twice or more...")
        selected[ic] = True
      if all(selected) and multi:
        output = self.newImage(f(self.image))
      else:
        output = self.newImage(np.empty_like(self.image))
        for ic in range(nc):
          if selected[ic]:
            output.image[ic] = f(self.image[ic])
          else:
            output.image[ic] =   self.image[ic]
      if trans: output.trans = transformation(f, self.image[selected], channels)
      return output

  def clip_channels(self, channels):
    """Clip selected channels of the image in the [0, 1] range.

    Args:
      channels (str): The selected channels:

        - An empty string (default): Apply the operation to all channels (RGB, HSV and grayscale images).
        - A combination of "1", "2", "3" (or equivalently "R", "G", "B" for RGB images):
          Apply the operation to the first/second/third channel (RGB, HSV and grayscale images).
        - "V": Apply the operation to the HSV value (RGB, HSV and and grayscale images).
        - "S": Apply the operation to the HSV saturation (RGB and HSV images).
        - "L": Apply the operation to the luma (RGB and grayscale images).
        - "Ls": Apply the operation to the luma, with highlights protection by desaturation.
          (after the operation, the out-of-range pixels are desaturated at constant luma).
        - "Lb": Apply the operation to the luma, with highlights protection by blending.
          (after the operation, the out-of-range pixels are blended with channels = "RGB").

    Returns:
      Image: The clipped image.
    """
    return self.apply_channels(lambda channel: np.clip(channel, 0., 1.), channels)

  def protect_highlights_sat(self):
    """Normalize out-of-range pixels with HSV value > 1 by adjusting the saturation at constant luma.

    The out-of-range RGB components of the pixels are decreased while the in-range RGB components are
    increased so that the luma is conserved. This desaturates (whitens) the pixels with out-ot-range
    components.
    This aims at protecting the highlights from overflowing when stretching the luma.

    Warning:
      The luma must be <= 1 even though some pixels have HSV value > 1.

    Returns:
      Image: The processed image.
    """
    self.check_color_model("RGB")
    imgluma = luma(self.image) # Original luma.
    if np.any(imgluma > 1.+params.IMGTOL/2):
      print("Warning, can not protect highlights if the luma is out-of-range. Returning original image...")
      return self.copy()
    newimage = self.image/np.maximum(self.image.max(axis = 0), 1.) # Rescale maximum HSV value to 1.
    newluma = luma(newimage) # Updated luma.
    # Scale the saturation.
    # Note: The following implementation is failsafe when newluma -> 1 (in which case luma is also 1 in principle),
    # at the cost of a small error.
    fs = ((1.-imgluma)+params.IMGTOL)/((1.-newluma)+params.IMGTOL)
    output = 1.-(1.-newimage)*fs
    diffluma = imgluma-luma(output)
    print(f"Maximum luma difference = {abs(diffluma).max()}.")
    return self.newImage(output)

  def protect_highlights_blend(self, bounded):
    """Normalize out-of-range pixels with HSV value > 1 by blending with a bounded image with HSV values <= 1.

    Each pixel of the image with out-of-range RGB components is brought back in-range by blending with the
    pixel of the input bounded image.
    This aims at protecting the highlights from overflowing when stretching the luma.

    Args:
      bounded (Image): The "in-range" image to blend with. All pixels must have HSV values <= 1.

    Returns:
      Image: The processed image.
    """
    self.check_color_model("RGB") ; bounded.check_color_model("RGB")
    if np.any(bounded.value() > 1.+params.IMGTOL/2):
      print("Warning, can not protect highlights if the input bounded image is out-of-range. Returning original image...")
      return self.copy()
    mixing = np.where(self.image > 1.+params.IMGTOL, helpers.failsafe_divide(self.image-1., self.image-bounded.image), 0.)
    return self.blend(bounded, mixing.max(axis = 0))



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.0.0 / 2024.12.15
# Sphinx OK.

"""Histogram stretch functions."""

import numpy as np

def shadow_stretch_function(x, shadow):
  """Return the linear shadow stretch function f(x).

  The input data x is clipped below shadow and linearly stretched
  to map [shadow, 1] onto [0, 1].
  The output, stretched data therefore fit in the [0, infty[ range.

  Args:
    x (numpy.ndarray): The input data.
    shadow (float): The shadow level (expected < 1).

  Returns:
    numpy.ndarray: The stretched data.
  """
  x = np.clip(x, shadow, None)
  return (x-shadow)/(1.-shadow)

def shadow_highlight_stretch_function(x, shadow, highlight):
  """Return the linear shadow/highlight stretch function f(x).

  The input data x is clipped below shadow and above highlight and linearly stretched
  to map [shadow, highlight] onto [0, 1].
  The output, stretched data therefore fit in the [0, 1] range.

  Args:
    x (numpy.ndarray): The input data.
    shadow (float): The shadow level (expected < 1).
    highlight (float): The highlight level (expected > shadow).

  Returns:
    numpy.ndarray: The stretched data.
  """
  x = np.clip(x, shadow, highlight)
  return (x-shadow)/(highlight-shadow)

def dynamic_range_stretch_function(x, fr, to):
  """Return the linear dynamic range stretch function f(x).

  The input data x is linearly stretched to map [fr[0], fr[1]] onto [to[0], to[1]],
  then clipped in the [to[0], to[1]] range.

  Args:
    x (numpy.ndarray): The input data.
    fr (a tuple or list of two floats): The input range.
    to (a tuple or list of two floats): The output range.

  Returns:
    numpy.ndarray: The stretched data.
  """
  return np.interp(x, fr, to)

def asinh_stretch_function(x, stretch):
  """Return the arcsinh stretch function f(x).

  The arcsinh stretch function is defined as:

    f(x) = arcsinh(stretch*x)/arcsinh(stretch)

  Args:
    x (numpy.ndarray): The input data.
    stretch (float): The stretch factor (expected >= 0).

  Returns:
    numpy.ndarray: The stretched data.
  """
  return np.arcsinh(stretch*x)/np.arcsinh(stretch) if abs(stretch) > 1.e-6 else x

def ghyperbolic_stretch_function(x, logD1, b, SYP, SPP, HPP, inverse):
  """Return the generalized hyperbolic stretch function f(x).

  For details about generalized hyperbolic stretches, see: https://ghsastro.co.uk/.
  This function clips the input data x in the [0, 1] range before stretching.

  Note:
  Code adapted from https://github.com/mikec1485/GHS/blob/main/src/scripts/GeneralisedHyperbolicStretch/lib/GHSStretch.js.

  Todo:
  Do not clip the input data and extend the transformation outside [0, 1] ?

  Args:
    x (numpy.ndarray): The input data.
    logD1 (float): The global stretch factor ln(D+1) (expected >= 0).
    b (float): The local stretch factor.
    SYP (float): The symmetry point (expected in [0, 1]).
    SPP (float): The shadow protection point (expected in [0, SYP]).
    HPP (float): The highlight protection point (expected in [SYP, 1]).
    inverse (bool): Return the inverse stretch function if True.

  Returns:
    numpy.ndarray: The stretched data.
  """
  D = np.exp(logD1)-1.
  x = np.clip(x, 0., 1.)
  if abs(D) < 1.e-6: # Identity.
    return x
  else:
    y = np.empty_like(x)
    if abs(b) < 1.e-6:
      qs = np.exp(-D*(SYP-SPP))
      q0 = qs-D*SPP*np.exp(-D*(SYP-SPP))
      qh = 2.-np.exp(-D*(HPP-SYP))
      q1 = qh+D*(1.-HPP)*np.exp(-D*(HPP-SYP))
      q  = 1./(q1-q0)
      # Coefficient for x < SPP.
      b1 = D*np.exp(-D*(SYP-SPP))*q
      # Coefficients for SPP <= x < SYP.
      a2 = -q0*q
      b2 = q
      c2 = -D*SYP
      d2 = D
      # Coefficients for SYP <= x < HPP.
      a3 = (2.-q0)*q
      b3 = -q
      c3 = D*SYP
      d3 = -D
      # Coefficients for x >= HPP.
      a4 = (qh-q0-D*HPP*np.exp(-D*(HPP-SYP)))*q
      b4 = D*np.exp(-D*(HPP-SYP))*q
      # GHS transformation.
      if not inverse:
        mask = (x <  SPP)
        y[mask] =    b1*x[mask]
        mask = (x >= SPP) & (x < SYP)
        y[mask] = a2+b2*np.exp(c2+d2*x[mask])
        mask = (x >= SYP) & (x < HPP)
        y[mask] = a3+b3*np.exp(c3+d3*x[mask])
        mask = (x >= HPP)
        y[mask] = a4+b4*x[mask]
      else:
        SPT =    b1*SPP
        SYT = a2+b2*np.exp(c2+d2*SYP)
        HPT = a4+b4*HPP
        mask = (x <  SPT)
        y[mask] = x[mask]/b1
        mask = (x >= SPT) & (x < SYT)
        y[mask] = (np.log((x[mask]-a2)/b2)-c2)/d2
        mask = (x >= SYT) & (x < HPT)
        y[mask] = (np.log((x[mask]-a3)/b3)-c3)/d3
        mask = (x >= HPT)
        y[mask] = (x[mask]-a4)/b4
    elif abs(b+1.) < 1.e-6:
      qs = -np.log(1.+D*(SYP-SPP))
      q0 = qs-D*SPP/(1.+D*(SYP-SPP))
      qh = np.log(1.+D*(HPP-SYP))
      q1 = qh+D*(1.-HPP)/(1.+D*(HPP-SYP))
      q  = 1./(q1-q0)
      # Coefficient for x < SPP.
      b1 = D/(1.+D*(SYP-SPP))*q
      # Coefficients for SPP <= x < SYP.
      a2 = -q0*q
      b2 = -q
      c2 = 1.+D*SYP
      d2 = -D
      # Coefficients for SYP <= x < HPP.
      a3 = -q0*q
      b3 = q
      c3 = 1.-D*SYP
      d3 = D
      # Coefficients for x >= HPP.
      a4 = (qh-q0-D*HPP/(1.+D*(HPP-SYP)))*q
      b4 = q*D/(1.+D*(HPP-SYP))
      # GHS transformation.
      if not inverse:
        mask = (x <  SPP)
        y[mask] =    b1*x[mask]
        mask = (x >= SPP) & (x < SYP)
        y[mask] = a2+b2*np.log(c2+d2*x[mask])
        mask = (x >= SYP) & (x < HPP)
        y[mask] = a3+b3*np.log(c3+d3*x[mask])
        mask = (x >= HPP)
        y[mask] = a4+b4*x[mask]
      else:
        SPT =    b1*SPP
        SYT = a2+b2*np.log(c2+d2*SYP)
        HPT = a4+b4*HPP
        mask = (x <  SPT)
        y[mask] = x[mask]/b1
        mask = (x >= SPT) & (x < SYT)
        y[mask] = (np.exp((x[mask]-a2)/b2)-c2)/d2
        mask = (x >= SYT) & (x < HPT)
        y[mask] = (np.exp((x[mask]-a3)/b3)-c3)/d3
        mask = (x >= HPT)
        y[mask] = (x[mask]-a4)/b4
    else:
      if b < 0.:
        b  = -b
        qs = (1.-(1.+D*b*(SYP-SPP))**((b-1.)/b))/(b-1.)
        q0 = qs-D*SPP*(1.+D*b*(SYP-SPP))**(-1./b)
        qh = ((1.+D*b*(HPP-SYP))**((b-1.)/b)-1.)/(b-1.)
        q1 = qh+D*(1.-HPP)*(1.+D*b*(HPP-SYP))**(-1./b)
        q  = 1./(q1-q0)
        # Coefficient for x < SPP.
        b1 = D*(1.+D*b*(SYP-SPP))**(-1./b)*q
        # Coefficients for SPP <= x < SYP.
        a2 = (1./(b-1.)-q0)*q
        b2 = -q/(b-1.)
        c2 = 1.+D*b*SYP
        d2 = -D*b
        e2 = (b-1.)/b
        # Coefficients for SYP <= x < HPP.
        a3 = (-1./(b-1.)-q0)*q
        b3 = q/(b-1.)
        c3 = 1.-D*b*SYP
        d3 = D*b
        e3 = (b-1.)/b
        # Coefficients for x >= HPP.
        a4 = (qh-q0-D*HPP*(1.+D*b*(HPP-SYP))**(-1./b))*q
        b4 = D*(1.+D*b*(HPP-SYP))**(-1./b)*q
      else:
        qs = (1.+D*b*(SYP-SPP))**(-1./b)
        q0 = qs-D*SPP*(1.+D*b*(SYP-SPP))**(-(1.+b)/b)
        qh = 2.-(1.+D*b*(HPP-SYP))**(-1./b)
        q1 = qh+D*(1.-HPP)*(1.+D*b*(HPP-SYP))**(-(1.+b)/b)
        q  = 1./(q1-q0)
        # Coefficient for x < SPP.
        b1 = D*(1.+D*b*(SYP-SPP))**(-(1.+b)/b)*q
        # Coefficients for SPP <= x < SYP.
        a2 = -q0*q
        b2 = q
        c2 = 1.+D*b*SYP
        d2 = -D*b
        e2 = -1./b
        # Coefficients for SYP <= x < HPP.
        a3 = (2.-q0)*q
        b3 = -q
        c3 = 1.-D*b*SYP
        d3 = D*b
        e3 = -1./b
        # Coefficients for x >= HPP.
        a4 = (qh-q0-D*HPP*(1.+D*b*(HPP-SYP))**(-(b+1.)/b))*q
        b4 = D*(1.+D*b*(HPP-SYP))**(-(b+1.)/b)*q
      # GHS transformation.
      if not inverse:
        mask = (x <  SPP)
        y[mask] =    b1*x[mask]
        mask = (x >= SPP) & (x < SYP)
        y[mask] = a2+b2*(c2+d2*x[mask])**e2
        mask = (x >= SYP) & (x < HPP)
        y[mask] = a3+b3*(c3+d3*x[mask])**e3
        mask = (x >= HPP)
        y[mask] = a4+b4*x[mask]
      else:
        SPT =    b1*SPP
        SYT = a2+b2*(c2+d2*SYP)**e2
        HPT = a4+b4*HPP
        mask = (x <  SPT)
        y[mask] = x[mask]/b1
        mask = (x >= SPT) & (x < SYT)
        y[mask] = (((x[mask]-a2)/b2)**(1./e2)-c2)/d2
        mask = (x >= SYT) & (x < HPT)
        y[mask] = (((x[mask]-a3)/b3)**(1./e3)-c3)/d3
        mask = (x >= HPT)
        y[mask] = (x[mask]-a4)/b4
    return y

def midtone_stretch_function(x, midtone, inverse):
  """Return the midtone stretch function f(x).

  The midtone stretch function is defined as:

    f(x) = (midtone-1)*x/((2*midtone-1)*x-midtone)

  In particular, f(0) = 0, f(midtone) = 0.5 and f(1) = 1.

  Args:
    x (numpy.ndarray): The input data.
    midtone (float): The midtone level (expected in ]0, 1[]).
    inverse (bool): Return the inverse stretch function if True.

  Returns:
    numpy.ndarray: The stretched data.
  """
  return midtone*x/((2.*midtone-1.)*x-midtone+1.) if inverse else (midtone-1.)*x/((2.*midtone-1.)*x-midtone)

def gamma_stretch_function(x, gamma):
  """Return the gamma stretch function f(x).

  The gamma stretch function is defined as:

    f(x) = x**gamma

  This function clips the input data x below 0 before stretching.

  Args:
    x (numpy.ndarray): The input data.
    gamma (float): The stretch exponent (expected > 0).

  Returns:
    numpy.ndarray: The stretched data.
  """
  x = np.clip(x, 0., None)
  return x**gamma

def midtone_levels_stretch_function(x, shadow, midtone, highlight, low, high):
  """Return the shadow/midtone/highlight/low/high levels adjustment function f(x).

  This function:

    1) Clips the input data in the [shadow, highlight] range and maps [shadow, highlight] to [0, 1].
    2) Applies the midtone stretch function f(x) = (m-1)*x/((2*m-1)*x-m),
       with m = (midtone-shadow)/(highlight-shadow) the remapped midtone.
    3) Maps [low, high] to [0, 1] and clips the output data in the [0, 1] range.

  Args:
    x (numpy.ndarray): The input data.
    midtone (float): The input midtone level (expected in ]0, 1[).
    shadow (float): The input shadow level (expected < midtone).
    highlight (float): The input highlight level (expected > midtone).
    low (float): The "low" output level (expected <= 0).
    high (float): The "high" output level (expected >= 1).

  Returns:
    numpy.ndarray: The stretched data.
  """
  midtone = (midtone-shadow)/(highlight-shadow)
  x = np.clip(x, shadow, highlight)
  x = (x-shadow)/(highlight-shadow)
  y = (midtone-1.)*x/((2.*midtone-1.)*x-midtone)
  return np.interp(y, (low, high), (0., 1.))



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.0.0 / 2024.12.15
# Sphinx OK.

"""Image geometry management."""

import numpy as np
from PIL import Image as PILImage

from . import params

#####################################
# For inclusion in the Image class. #
#####################################

class Mixin:
  """To be included in the Image class."""

  ################################
  # Geometrical transformations. #
  ################################

  def flip_height(self):
    """Flip the image along its height.

    Returns:
      Image: The flipped image.
    """
    return np.flip(self, axis = 1)

  def flip_width(self):
    """Flip the image along its width.

    Returns:
      Image: The flipped image.
    """
    return np.flip(self, axis = 2)

  ##################
  # Resize & Crop. #
  ##################

  def resize(self, width, height, method = "lanczos"):
    """Resize the image.

    Args:
      width (int): New image width (pixels).
      height (int): New image height (pixels).
      method (str, optional): Resampling method:

        - "nearest": Nearest neighbor interpolation.
        - "bilinear": Linear interpolation.
        - "bicubic": Cubic spline interpolation.
        - "lanczos": Lanczos (truncated sinc) filter (default).
        - "box": Box average (equivalent to "nearest" for upscaling).
        - "hamming": Hamming (cosine bell) filter.

    Returns:
      Image: The resized image.
    """
    if width < 1 or width > 32768: raise ValueError("Error, width must be >= 1 and <= 32768 pixels.")
    if height < 1 or height > 32768: raise ValueError("Error, height must be >= 1 and <= 32768 pixels.")
    if width*height > 2**26: raise ValueError("Error, can not resize to > 64 Mpixels.")
    if method == "nearest":
      method = PILImage.Resampling.NEAREST
    elif method == "bilinear":
      method = PILImage.Resampling.BILINEAR
    elif method == "bicubic":
      method = PILImage.Resampling.BICUBIC
    elif method == "lanczos":
      method = PILImage.Resampling.LANCZOS
    elif method == "box":
      method = PILImage.Resampling.BOX
    elif method == "hamming":
      method = PILImage.Resampling.HAMMING
    else:
      raise ValueError(f"Error, unknown resampling method '{method}'.")
    nc = self.get_nc()
    output = np.empty((nc, height, width), dtype = params.IMGTYPE)
    for ic in range(nc): # Resize each channel using PIL.
      PILchannel = PILImage.fromarray(np.float32(self.image[ic]), "F").resize((width, height), method) # Convert to np.float32 while resizing.
      output[ic] = np.asarray(PILchannel, dtype = params.IMGTYPE)
    return self.newImage(output)

  def rescale(self, scale, method = "lanczos"):
    """Rescale the image.

    Args:
      scale (float): Scaling factor.
      method (str, optional): Resampling method:

        - "nearest": Nearest neighbor interpolation.
        - "bilinear": Linear interpolation.
        - "bicubic": Cubic spline interpolation.
        - "lanczos": Lanczos (truncated sinc) filter (default).
        - "box": Box average (equivalent to "nearest" for upscaling).
        - "hamming": Hamming (cosine bell) filter.

    Returns:
      Image: The rescaled image.
    """
    if scale <= 0. or scale > 16.: raise ValueError("Error, scale must be > 0 and <= 16.")
    width, height = self.get_size()
    newwidth, newheight = int(round(scale*width)), int(round(scale*height))
    return self.resize(newwidth, newheight, method)

  def crop(self, xmin, xmax, ymin, ymax):
    """Crop the image.

    Args:
      xmin, xmax, ymin, ymax (integer or float): Crop from x = xmin to x = xmax and from y = ymin to y = ymax.

    Returns:
      Image: The cropped image.
    """
    if xmax <= xmin: raise ValueError("Error, xmax <= xmin.")
    if ymax <= ymin: raise ValueError("Error, ymax <= ymin.")
    width, height = self.get_size()
    xmin = max(int(np.floor(xmin))  , 0)
    xmax = min(int(np.ceil (xmax))+1, width)
    ymin = max(int(np.floor(ymin))  , 0)
    ymax = min(int(np.ceil (ymax))+1, height)
    return self.newImage(self.image[:, ymin:ymax, xmin:xmax])



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.0.0 / 2024.12.15

"""Launcher for eQuimageLab."""

import os
import sys
import shutil
import tkinter
import tkinter.filedialog
import tkinter.messagebox
import subprocess
from pathlib import Path
from PIL import Image, ImageTk

def run():
  """Run eQuimageLab launcher."""

  def new_notebook():
    """Create and open a new jupyter notebook."""
    # Open file selection dialog.
    notebook = tkinter.filedialog.asksaveasfilename(title = "Save new notebook as", filetypes = [("Jupyter notebooks", "*.ipynb")],
                                                    initialdir = Path.home(), initialfile = "eqlab.ipynb", defaultextension = ".ipynb")
    if not notebook: return
    # Create notebook.
    template = os.path.join(__packagepath__, "launcher", "equimagelab.ipynb")
    print(f"Copying {template} as {notebook}...")
    try:
      shutil.copyfile(template, notebook)
    except Exception as err:
      tkinter.messagebox.showerror("Error", f"Failed to create notebook {notebook}:\n{str(err)}")
      return
    # Run jupyter lab.
    run_jupyter_lab(notebook)

  def open_notebook():
    """Select and open a jupyter notebook."""
    # Open file selection dialog.
    notebook = tkinter.filedialog.askopenfilename(title = "Open notebook", filetypes = [("Jupyter notebooks", "*.ipynb")], initialdir = Path.home())
    if not notebook: return
    # Run jupyter lab.
    run_jupyter_lab(notebook)

  def open_directory():
    """Run jupyter lab in a selected directory."""
    # Open directory selection dialog.
    directory = tkinter.filedialog.askdirectory(title = "Open directory", initialdir = Path.home())
    if not directory: return
    # Run jupyter lab.
    run_jupyter_lab(f"--notebook-dir={directory}")

  def run_jupyter_lab(options = ""):
    """Run jupyter lab.

    Args:
      options (str): Options for the jupyter lab command (default "").
    """
    print(f"Running jupyter lab {options}...")
    try:
      subprocess.Popen(["jupyter", "lab", options])
    except Exception as err:
      tkinter.messagebox.showerror("Error", f"Failed to run jupyter lab:\n{str(err)}")
      return
    quit()

  def quit():
    """Quit launcher."""
    root.destroy()
    sys.exit(0)

  from .. import __packagepath__
  # Open root Tk window.
  root = tkinter.Tk()
  root.title("eQuimageLab")
  # Configure menu.
  menu = tkinter.Menu(root)
  menu.add_command(label = "New notebook", command = new_notebook)
  menu.add_command(label = "Open notebook", command = open_notebook)
  menu.add_command(label = "Open directory", command = open_directory)
  menu.add_command(label = "Quit", command = quit)
  root.config(menu = menu)
  # Display splash screen.
  canvas = tkinter.Canvas(root, width = 800, height = 600)
  canvas.pack(side = "top")
  try:
    image = Image.open(os.path.join(__packagepath__, "images", "splash.png")).resize((800, 600))
  except:
    pass
  else:
    imagetk = ImageTk.PhotoImage(image)
    canvas.create_image(0, 0, anchor = "nw", image = imagetk)
  # Start Tk main loop.
  root.mainloop()
